/* calc.c */

#include <stdio.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include "waved.h"
#include "video.h"
#include "memory.h"

#define MAXNFM  (MAXLPC/2)
#define NFFT	512
#ifndef PI
#define PI      3.1415926535898
#endif

char   *marktext();
int     autoc();
int     findpk();
int     rfft();
int     menu();
int     ib_open();
void    blankmarks();
void    calc_rms();
void    calc_durs();
void    calc_lpc();
void    calc_formant();
void    calc_moments();
void    ordmrk();
void    beep();
void    g_box();
void    g_drawmark();
void    g_rstscr();
void    get_data(float *, SIGPTR, int, int, int, int);
void    ib_text();
void    ib_close();

void
calcquery()
{
    int     c;
    static char *mis[] = {
    	"Area", "Duration", "Formant", "LPC", "Moments", "RMS", ""
    };
    
    c = menu("CALCULATE", mis);
    switch (c) {
    case 'A':
	calc_lpc(1);
	break;
    case 'D':
	calc_durs();
	break;
    case 'F':
	calc_formant();
	break;
    case 'L':
	calc_lpc(0);
	break;
    case 'M':
	calc_moments();
	break;
    case 'R':
	calc_rms();
	break;
    }
}

void
calc_rms()
{
    char    str[80];
    double  val, sum, avg, rms;
    int     zer, vmn, vmx;
    long    aaa, bbb, i, n;
    SIGPTR  buf;

    if (getwinmrk(" RMS:", 0) == 0) {
	return;
    }
    ordmrk();
    aaa = sig[aw].sigmrk[seg_mark1];
    bbb = sig[aw].sigmrk[seg_mark2];
    clrmsg();
    putmsg("Computing signal RMS value ...");
    n = bbb - aaa;
    buf = sig[aw].sigbuf;
    zer = sig[aw].sigzer;
    sum = 0;
    for (i = aaa; i < bbb; i++) {
	val = buf[i] - zer;
	sum += val;
    }
    avg = sum / n;
    sum = 0;
    vmn = vmx = buf[aaa];
    for (i = aaa; i < bbb; i++) {
	val = (buf[i] - zer) - avg;
	sum += val * val;
	if (vmn > buf[i])
	    vmn = buf[i];
	if (vmx < buf[i])
	    vmx = buf[i];
    }
    rms = sqrt(sum / n);

    sprintf(str, " RMS = %.1f,  DC = %.1f,  MIN = %d,  MAX = %d,  NSAMP = %ld.",
	rms, avg, vmn, vmx, n);
    putmsgwk(str);
}

/* CALC_DURS to give the time between two markers */

void
calc_durs()
{
    char    str[80], mt1[10], mt2[10], *u;
    double  t;
    long    aaa, bbb, n;
    SIGNAL *s;

    s = &sig[aw];
    if (s == NULL) {
	putmsgwk("No signal.");
	return;
    }
    if (s->adrate <= 0) {
	putmsgwk(" Undefined signal ADC rate.");
	return;
    }
    if (getwinmrk(" DURATION:", 0) == 0) {
	return;
    }
    ordmrk();
    aaa = s->sigmrk[seg_mark1];
    bbb = s->sigmrk[seg_mark2];
    n = (aaa < bbb) ? bbb - aaa : aaa - bbb;
    t = n * 1000.0 / s->adrate;
    u = "msec";
    if (t >= 1e4) {
        t /= 1e3;
        u = "sec";
    }
    strcpy(mt1, marktext(seg_mark1));
    strcpy(mt2, marktext(seg_mark2));
    sprintf(str, " Duration from %s to %s = %9.3f %s, (%ld samples).", 
        mt1, mt2, t, u, n);
    putmsgwk(str);
}

/* calc_lpc - calculate linear prediction coefficients */

void
calc_lpc(aflg)
{
    char    str[80];
    int     nrow, ncol, xw, yh, xx, yy, cw;
    int     nc, i, j, np, status, ncp, zer;
    float  *x, *a, *rc, *r, alpha, sm, refl;
    long    aaa, bbb, npt;
    SIGPTR  buf;
    SIGNAL *s;
    extern int lpnco, lpwnt, lphfe;

    if (getwinmrk(" LPC:", 0) == 0) {
	return;
    }
    ordmrk();
    clrmsg();
    s = &sig[aw];
    if (s == NULL) {
	putmsgwk("No signal.");
	return;
    }
    aaa = s->sigmrk[seg_mark1];
    bbb = s->sigmrk[seg_mark2];
    if (aaa >= bbb) {
	putmsgwk("No points in segment.");
	return;
    }
    zer = s->sigzer;
    buf = s->sigbuf + aaa;
    npt = bbb - aaa;
    if (npt > 32767) {
	putmsgwk("Too many points in segment.");
	return;
    }
    np = npt;
    nc = (lpnco > 0) ? lpnco : 2 + s->adrate / 1000;
    if (nc <= 0 || nc > MAXLPC || nc > np) {
	putmsgwk("Invalid number of coefficients.");
	return;
    }
    putmsg("computing linear prediction coefficients ...");
    x = (float *) MALLOC(np, sizeof(float));
    if (x == NULL) {
	putmsgwk("Problem allocating memory for FFT.");
	return;
    }
    ncp = nc + 1;
    rc = (float *) MALLOC(ncp * 3, sizeof(float));
    if (rc == NULL) {
	putmsgwk("Problem allocating memory for LPC.");
	FREE(x);
	return;
    }
    a = rc + ncp;
    r = a + ncp;
    get_data(x, buf, np, zer, lphfe, lpwnt);
    status = autoc(np, x, nc, a, &alpha, rc, r);
    clrmsg();
    if (status != 0) {
	putmsgwk("Singular autocorrelation matrix.");
	FREE(x);
	FREE(rc);
	return;
    }
    nrow = 14;
    ncol = (nc <= 30) ? 3 : (nc + 9) / 10;
    cw = 15;
    xw = (ncol * cw + 2) * ch_wd;
    yh = (nrow + 2) * ln_sp;
    if (ib_open(xw, yh) < 0) {
	FREE(x);
	FREE(rc);
	return;
    }
    xx = 15;
    yy = 25;
    if (aflg == 0) {
	ib_text(xx, yy, "Linear Prediction Coefficients:");
	for (i = 1; i <= nc; i++) {
	    j = i - 1;
	    sprintf(str, "%4d%8.3f", i, a[i]);
	    ib_text(xx + cw * (j / 10) * ch_wd, yy + (2 + j % 10) * ln_sp, str);
	}
    } else {
	a[0] = sm = 1.0;
	for (i = 1; i <= nc; i++) {
	    refl = rc[nc - i];
	    sm *= (float) ((1.0 + refl) / (1.0 - refl));
	    a[i] = sm;
	}
	ib_text(xx, yy, "Acoustic Tube Section Areas:");
	for (i = 1; i <= nc; i++) {
	    j = i - 1;
	    sprintf(str, "%4d%8.2f", i, a[i]);
	    ib_text(xx + cw * (j / 10) * ch_wd, yy + (2 + j % 10) * ln_sp, str);
	}
    }
    FREE(x);
    FREE(rc);
    (void) getevent();
    ib_close();
}

/* calc_formant - calculate formants */

void
calc_formant()
{
    char    str[80];
    int     nrow, ncol, xw, yh, xx, yy, cw;
    int     nfft, nfrq, j, nf = 0;
    int     nc, i, np, status, ncp, zer;
    float  *x, *a, *rc, *r, alpha;
    float   fm[MAXNFM], bw[MAXNFM];
    float   xi, xr, dn;
    long    aaa, bbb, npt;
    SIGPTR  buf;
    SIGNAL *s;
    extern int lpnco, lpwnt, lphfe;

    if (getwinmrk(" FORMANT:", 0) == 0) {
	return;
    }
    ordmrk();
    clrmsg();
    s = &sig[aw];
    if (s == NULL) {
	putmsgwk("No signal.");
	return;
    }
    aaa = s->sigmrk[seg_mark1];
    bbb = s->sigmrk[seg_mark2];
    if (aaa >= bbb) {
	putmsgwk("No points in segment.");
	return;
    }
    zer = s->sigzer;
    buf = s->sigbuf + aaa;
    npt = bbb - aaa;
    if (npt > 32767) {
	putmsgwk("Too many points in segment.");
	return;
    }
    np = npt;
    nc = (lpnco > 0) ? lpnco : 2 + s->adrate / 1000;
    if (nc <= 0 || nc > MAXLPC || nc > np) {
	putmsgwk("Invalid number of coefficients.");
	return;
    }
    putmsg("computing linear prediction coefficients ...");
    ncp = nc + 1;
    rc = (float *) MALLOC(ncp * 3, sizeof(float));
    if (rc == NULL) {
	putmsgwk("Problem allocating memory for LPC.");
	return;
    }
    a = rc + ncp;
    r = a + ncp;
    for (nfft = 2; nfft < np; nfft *= 2)
    	continue;
    x = (float *) MALLOC(nfft + 2, sizeof(float));
    if (x == NULL) {
	putmsgwk("Problem allocating memory for FFT.");
	FREE(x);
	return;
    }
    get_data(x, buf, np, zer, lphfe, lpwnt);
    status = autoc(np, x, nc, a, &alpha, rc, r);
    if (status != 0) {
	putmsgwk("Singular autocorrelation matrix.");
	FREE(x);
	FREE(rc);
	return;
    }
    for (i = 0; i < nc; i++)
        x[i] = a[i];
    for (i = nc; i < nfft; i++)
        x[i] = 0.0;
    (void) rfft(x, nfft);
    nfrq = nfft / 2;
    for (i = 0; i < nfrq; i++) {	/* complex magnitude squared */
        j = 2 * i;
        xr = x[j];
        xi = x[j + 1];
        dn = (xr * xr + xi * xi);
        x[i] = (dn > 0.0) ? alpha / dn : alpha;
    }
    nf = findpk(x, nfrq, (float) s->adrate, fm, bw);

    clrmsg();
    nrow = 14;
    ncol = (nf < 20) ? 2 : (nf + 9) / 10;
    cw = 20;
    xw = (ncol * cw + 2) * ch_wd;
    yh = (nrow + 2) * ln_sp;
    if (ib_open(xw, yh) < 0) {
	FREE(x);
	FREE(rc);
	return;
    }
    xx = 15;
    yy = 25;
    ib_text(xx, yy, "Formant: frequency / bandwidth (Hz)");
    for (i = 1; i <= nf; i++) {
	j = i - 1;
	sprintf(str, "%4d%6.0f / %.0f", i, fm[j], bw[j]);
	ib_text(xx + cw * (j / 10) * ch_wd, yy + (2 + j % 10) * ln_sp, str);
    }
    FREE(x);
    FREE(rc);
    (void) getevent();
    ib_close();
}

void
calc_moments()
{
    char    str[80];
    double  wt = 0, m1 = 0, m2 = 0, m3 = 0, m4 = 0;
    double  sc, df, wf, sd, dp, cp, ph, wn, n3m, n4m;
    int     nfft, nfrq, j;
    int     i, np;
    float  *x;
    float   xi, xr;
    long    aaa, bbb, npt;
    SIGPTR  buf;
    SIGNAL *s;
    extern int fmwnt, fmhfe;
    static double tpi = 2.0 * 3.1415926535898;
    static double hwc[3] = {0.53836, 0.46164}; /* Hamming window coefficients */
    static double bwc[3] = {0.42, 0.5, 0.08}; /* Blackman window coefficients */

    if (getwinmrk(" MOMENTS:", 0) == 0) {
	return;
    }
    ordmrk();
    clrmsg();
    s = &sig[aw];
    if (s == NULL) {
	putmsgwk("No signal.");
	return;
    }
    aaa = s->sigmrk[seg_mark1];
    bbb = s->sigmrk[seg_mark2];
    if (aaa >= bbb) {
	putmsgwk("No points in segment.");
	return;
    }
    buf = s->sigbuf + aaa;
    npt = bbb - aaa;
    if (npt > 16384) {
	putmsgwk("Too many points in segment.");
	return;
    }
    if (npt < 2) {
	putmsgwk("Too few points in segment.");
	return;
    }
    putmsg("computing moments ...");
    np = npt;
    for (nfft = 2; nfft < np; nfft *= 2)
    	continue;
    x = (float *) MALLOC(nfft, sizeof(float));
    if (x == NULL) {
	beep();
	putmsgwk("Problem allocating memory!");
	return;
    }
    dp = tpi / (np - 1);
    cp = (np - 1) / 2.0;
    for (i = 0; i < np; i++) {
	ph = (i - cp) * dp;
        if (fmwnt == 0)
            wn = 1;
        else if (fmwnt == 1)
            wn = hwc[0] + hwc[1] * cos(ph);
        else
            wn = bwc[0] + bwc[1] * cos(ph) + bwc[2] * cos(2 * ph);
        if (fmhfe == 0)
            x[i] = (float) (buf[i] * wn);
        else if (i == 0)
            x[i] = 0;
        else
            x[i] = ((float) buf[i] - (float) buf[i - 1]) * (float) wn;
    }
    for (i = np; i < nfft; i++)
        x[i] = 0.0;
    (void) rfft(x, nfft);
    nfrq = nfft / 2;
    sc = (0.001 * s->adrate) / nfft;
    for (i = 0; i < nfrq; i++) {	/* complex magnitude squared */
        j = 2 * i;
        xr = x[j];
        xi = x[j + 1];
        x[i] = (xr * xr + xi * xi);
        wt += x[i];
        m1 += i * sc * x[i];
    }
    m1 /= wt;
    for (i = 0; i < nfrq; i++) {
        df = (i * sc - m1);
        wf = df * df * x[i];
        m2 += wf;
        wf *= df;
        m3 += wf;
        wf *= df;
        m4 += wf;
    }
    m2 /= wt;
    m3 /= wt;
    m4 /= wt;
    FREE(x);

    sd = sqrt(m2);
    n3m = m3 / (sd * sd * sd);
    n4m = (m4 / (m2 * m2)) - 3;
    sprintf(str, " Moments:  "
        "mean=%.3f(kHz) sd=%.3f(kHz) n3m=%.3f n4m=%.3f", m1, sd, n3m, n4m);
    putmsgwk(str);
}
