/* device.c */

#include <dos.h>
#include "waved.h"
#include "video.h"
#include "iod.h"

#define DEV_INTERNAL    0
#define DEV_PORT        1
#define DEV_DT          2
#define DEV_TDT2        3
#define DEV_MON         4
#define DEV_PIN         5

int     menu();
int     init_iod();
int     init_iod0();
int     init_iod1();
int     init_iod2();
int     init_iod3();
int     init_iod4();
int     init_iod5();

static void portquery();
static void dtquery();
static void tbquery();

void
devicequery()
{
    char    msg[80];
    int     c, err;
    static char *mis[] = {
    	"Internal", "DT", "TDT2", "Monterey", "Pinnacle", ""
    };
    
    c = menu("DEVICE", mis);    	

    switch (c) {
    case 'M':
	iod_num = DEV_MON;
	tbquery();
	break;
    case 'D':
	iod_num = DEV_DT;
	dtquery();
	break;
    case 'I':
	iod_num = DEV_INTERNAL;
	break;
    case 'P':
	iod_num = DEV_PIN;
	tbquery();
	break;
    case 'T':
	iod_num = DEV_TDT2;
	break;
    default:
        return;
    }
    stop_iod();
    err = init_iod();
    if (err != 0) {
	sprintf(msg, " Error in 'init_iod', error code = %d ", err);
    } else {
	if (iod_num == DEV_INTERNAL) {
	    sprintf(msg, " I/O Device = PC internal speaker");
	} else if (iod_num == DEV_PORT) {
	    sprintf(msg, " I/O Device = 8-bit DAC at port #%x", iod_port);
	} else if (iod_num == DEV_TDT2) {
	    sprintf(msg, " I/O Device = TDT2");
	} else if (iod_num == DEV_DT) {
	    sprintf(msg,
		" I/O Device = DT:  12-bits (unsigned), AD gain=%d, AD chan=%x, DA chan=%d",
		1 << adgain, adchan, dachan);
	} else if (iod_num == DEV_MON) {
	    sprintf(msg, " I/O Device = Monterey/Tahiti");
	} else if (iod_num == DEV_PIN) {
	    sprintf(msg, " I/O Device = Pinnacle/Fiji");
	}
    }
    putmsgwk(msg);
}

static void
dtquery()
{
    int     c;
    static char *mis1[] = {"1", "2", "3", "4", "5", ""};
    static char *mis2[] = {"0", "1", ""};
    static char *mis3[] = {
        "0", "1", "2", "3", "4", "5", "6", "7",
        "8", "9", "A", "B", "C", "D", "E", "F", ""
    };

    c = menu("DT GAIN", mis1);
    if (c == '1') {
	adgain = 0;
    } else if (c == '2') {
	adgain = 1;
    } else if (c == '4') {
	adgain = 2;
    } else if (c == '8') {
	adgain = 3;
    }

    c = menu("DAC CHANNEL", mis2);
    if (c == '0') {
	dachan = 0;
    } else if (c == '1') {
	dachan = 1;
    }

    c = menu("ADC CHANNEL", mis3);
    if (c >= '0' && c <= '9') {
	adchan = c - '0';
    } else if (c >= 'A' && c <= 'F') {
	adchan = c - 'A' + 10;
    }
}

static void
tbquery()
{
    putmsg(" Enable audio feedthrough? (y/N) ");
    adfeed = getyes();
}

void
find_iod()
{
    char    msg[80];
    static char *iod_str[] = {
        "Internal", "Port", "DT", "TDT", "Monterey", "Pinnacle"
    };

    if (init_iod5() == 0) {
        iod_num = 5;
    } else if (init_iod4() == 0) {
        iod_num = 4;
    } else if (init_iod3() == 0) {
        iod_num = 3;
    } else if (init_iod2() == 0) {
        iod_num = 2;
    } else {
        (void) init_iod0();
        iod_num = 0;
    }
    sprintf(msg, " Found I/O device = %s", iod_str[iod_num]);
    putmsgwk(msg);
}

int
init_iod()
{
    if (iod_num < 0)
        find_iod();
    switch (iod_num) {
    case DEV_INTERNAL:
        return (init_iod0());
    case DEV_PORT:
        return (init_iod1());
    case DEV_DT:
        return (init_iod2());
    case DEV_TDT2:
        return (init_iod3());
    case DEV_MON:
        return (init_iod4());
    case DEV_PIN:
        return (init_iod5());
    }
    return (2);
}
