/* file.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <direct.h>
#include "waved.h"
#include "video.h"
#include "file.h"
#include "memory.h"
#include "fk.h"
#ifdef WIN32
#include "windows.h"
#endif /* WIN32 */
#ifdef DJC
#include <unistd.h>
#define _chdir chdir
#endif

#define ESC 27

int     fillbuff();
int     getevent();
int     poscurs();
int     open_output();
int     ratecode();
int     menu();
void    leave();
void    loadfile();
void    savefile();
void    scan_signal();
void    sigmsg();
void    rootfn();
void    showall();

char   adf_key[8] = "CSRE40";
char   *btni_key = "BTNI_";
long   snd_key = 0x2E736E64L;
char   *trn_key = "TRN";
char   *ft[NFT] ={
    "ILS", "C", "TRN", "LST", "PLT", "SND", "NOH", "MAT", 
    "WAV", "TOK", "ADF"
    };
char    bps[8] = {1, 1, 1, 2, 3, 4, 4, 8};
int     file_type = FT_ILS;
int     data_type = DT_LINEAR_16;
int     byte_swap = 0;
long    cust_head = 0;
long    cust_tail = 0;

static void savefileseg();
static void save_as();
static void custom_load();

void
filequery()
{
    int     c;
    static char *mis[] =
    {
	"Load", "Reload", "Custom_load", "Save", "save_As", "eXit", ""
    };

    c = menu("FILE", mis);
    if (c == 'L') {
	loadfile("", 0);
    } else if (c == 'R') {
	loadfile(sig[aw].sigfnm, 0);
    } else if (c == 'S') {
	savefileseg();
    } else if (c == 'A') {
	save_as();
	savefileseg();
    } else if (c == 'C') {
	custom_load();
	loadfile("", 1);
    } else if (c == 'X') {
	putmsg(" Are you sure you want to EXIT? (y/N) ");
	if (getyes()) {
	    leave(0);
	} else {
	    clrmsg();
	}
    }
    clrmsg();
}

static void
savefileseg()
{
    long    p1, p2;
    SIGNAL *s;

    s = &sig[aw];
    if (s->siglen == 0)
	return;
    if (getwinmrk(" FILE SAVE:", 0)) {
	p1 = s->sigmrk[seg_mark1];
	p2 = s->sigmrk[seg_mark2];
	savefile(p1, p2);
    }
}

static void
save_as()
{
    int     c;
    static char *mis[] =
    {
	"ADF", "C", "ILS", "LST", "PLT", "SND", "MAT", "WAV", "NOH", ""
    };

    c = menu("SAVE FILE TYPE", mis);
    switch (c) {
    case 'A':
	file_type = FT_ADF;
	break;
    case 'C':
	file_type = FT_C;
	break;
    case 'I':
	file_type = FT_ILS;
	break;
    case 'L':
	file_type = FT_LST;
	break;
    case 'P':
	file_type = FT_PLT;
	break;
    case 'S':
	file_type = FT_SND;
	break;
    case 'M':
	file_type = FT_MAT;
	break;
    case 'W':
	file_type = FT_WAV;
	break;
    case 'N':
	file_type = FT_NOH;
	break;
    }
}

static void
custom_load()
{
    char    msg[80];
    int     c, b;
    static char *mis1[] = {"1", "2", "4", "8", ""};
    static char *mis2[] = {"0-unsigned", "1-signed", "2-mulaw", ""};

    sprintf(msg, " File Header size [%ld bytes] ?  ", cust_head);
    get_response(msg, msg, 80);
    cust_head = atol(msg);

    sprintf(msg, " File Trailer size [%ld bytes] ?  ", cust_tail);
    get_response(msg, msg, 80);
    cust_tail = atol(msg);

    b = (data_type < 3) ? 1 : 2;
    sprintf(msg, "Bytes per sample [%d]", b);
    c = menu(msg, mis1);
    switch (c) {
    case '1':
	sprintf(msg, "Data type [%d]", data_type);
	c = menu(msg, mis2);
	switch (c) {
	case '0':
	    data_type = DT_UNSIGNED_8;
	    break;
	case '1':
	    data_type = DT_LINEAR_8;
	    break;
	case '2':
	    data_type = DT_MULAW_8;
	    break;
	}
	break;
    case '2':
	data_type = DT_LINEAR_16;
	break;
    case '4':
	putmsg(" Floating-point data (y/N) ? ");
        data_type = getyes() ? DT_FLOAT : DT_LINEAR_32;
	break;
    case '8':
	data_type = DT_DOUBLE;
	break;
    default:
	data_type = DT_LINEAR_16;
    }
    if (c == '2' || c == '4' || c == '8') {
	putmsg(" Swap bytes (y/N) ? ");
	byte_swap = getyes();
    } else {
        byte_swap = 0;
    }
}

/****************************************************************************/

void
scan_signal(int sn)
{
    int     smn, smx, ss, bits, zero;
    long    sum, range, average, i, n;
    SIGPTR  buf;
    SIGNAL *s;

    s = &sig[sn];
    buf = s->sigbuf;
    n = s->siglen * s->sgnchn;
    sum = 0;
    smn = smx = buf[0];
    for (i = 1; i < n; i++) {
	ss = buf[i];
	if (ss > smx)
	    smx = ss;
	if (ss < smn)
	    smn = ss;
	sum += ss;
    }
    average = (n > 0) ? sum / n : 0;
    range = ((long) smx) - ((long) smn);
    if (range > 4096 || (average < 1024 && smx >= 2048)) {
	bits = 16;
	zero = 0;
    } else if (range > 256 || average > 256) {
	bits = 12;
	zero = (average < 1024) ? 0 : 2048;
    } else {
	bits = 8;
	zero = (average < 64) ? 0 : 128;
    }

    s->sgbits = bits;
    s->sigzer = zero;
    s->sigmin = smn;
    s->sigmax = smx;
}

void
sigmsg(int wn)
{
    char    msg[80];
    SIGNAL *s;

    if (siginfmsg == 0)		/* suppress signal info messages? */
	return;
    clrmsg();
    s = &sig[wn];
    if (s->siglen == 0) {
	putmsg("No signal.");
	return;
    }
    sprintf(msg,
	" (%s) %ld samples (%d chan.) at %ld Hz, ADC = %d bits (%ssigned) ",
	ft[(int)s->sigfty], s->siglen, s->sgnchn, s->adrate, s->sgbits,
	(s->sigzer == 0) ? "" : "un");
    putmsgwk(msg);
}

void
rootfn(char *rfn, char *fn)
{
    char   *h, *t, *s;

    for (h = fn; (s = strchr(h, '\\')) != NULL; h = s)
	continue;
    for (; (s = strchr(h, '/')) != NULL; h = s)
	continue;
    if ((s = strchr(h, '.')) != NULL)
	t = s;
    else
	t = h + strlen(h);
    for (s = h; s < t; s++)
	*rfn++ = *s;
    *rfn = '\0';
}

static long
exchange(long next)
{
    unsigned char *ch, chtmp;

    ch = (unsigned char *) &next;
    chtmp = ch[0];
    ch[0] = ch[3];
    ch[3] = chtmp;
    chtmp = ch[1];
    ch[1] = ch[2];
    ch[2] = chtmp;
    return (next);
}

void
exch_snd_hdr(sound)
SND_fmt *sound;
{
    sound->key = exchange(sound->key);
    sound->ofst = exchange(sound->ofst);
    sound->size = exchange(sound->size);
    sound->code = exchange(sound->code);
    sound->rate = exchange(sound->rate);
    sound->chan = exchange(sound->chan);
}

#define	MAXNF	80
#define	MAXSL	80
#define	FL	(FNMSZ+1)

void
trim(char *s)
{
    char *e;

    e = s + strlen(s);
    while ((e > s) && (e[-1] == ' '))
	e--;
    *e = '\0';
}

int
change_dir(char *s)
{
#ifdef WIN32
    if (strlen(s) > 2 && s[1] == ':') {
	if (s[0] >= 'A' && s[0] <= 'Z')
	    _chdrive(s[0] - 'A' + 1);
	if (s[0] >= 'a' && s[0] <= 'z')
	    _chdrive(s[0] - 'a' + 1);
	s += 2;
    }
#endif /* WIN32 */
    return (_chdir(s));
}

void
cdcpfn(char *fn, char *path)
{
    char *cd, *s;
	int  n;

    n = strlen(path);
    cd = (char *) malloc(n + 1);
    strcpy(cd, path);
    for (s = cd + n - 1; s > cd; s--) {
    	if(*s == '/' || *s == '\\' || *s == ':') {
	    s++;
	    *s = '\0';
	    change_dir(cd);
		break;
	}
    }
    strncpy(fn, path + (s - cd), FL);
    fn[FL - 1] = '\0';
    free(cd);
}

void
get_file_name(prmpt, fn, nc)
char   *prmpt, *fn;
{
    char   *filnam, *fs, *s, cd[PATHSZ], ncd[PATHSZ], ds[MAXSL], nm[16];
    int     yy, done = 0, select, lstsel, xs, ys, cs, lxs = 0, lys = 0;
    int     w, h, c, r, i, x, y, nf, mx, my, tc = 0, wc = 0, diroff = 0;
    int     hh, cc = 16, y0 = 4, x0 = 1, cn = 4, rn = 16;
    int     ibx1, iby1, ibx2, iby2, ibss;
#ifdef WIN32
    HANDLE  hnd;
    WIN32_FIND_DATA f;
#else  /* WIN32 */
    struct _find_t f;
#endif /* WIN32 */
    static char *errmsg = "Problem opening filename box 1.";

    clrmsg();
    filnam = (char *) MALLOC(MAXNF, FL);
    if (filnam == (char *) NULL)
	return;
    c = ch_wd;
    r = ln_sp;
    w = (cn * cc + 2) * ch_wd;
    h = y0 * r;
    hh = y0 + rn;
    ibx1 = (scrszx - w) / 2;
    iby1 = (scrszy - h) / 2;
    ibx2 = ibx1 + w - 1;
    iby2 = iby1 + h - 1;
    ibss = g_savscr(ibx1, iby1, ibx2, iby2);
    if (ibss < 0) {
	putmsgwk(errmsg);
	return;
    } else {
	g_bar(ibx1, iby1, ibx2, iby2, hsbgc);
	g_box(ibx1 + 1, iby1 + 1, ibx2 - 1, iby2 - 1, hsfgc);
    }
    if (*fn == '[')
	*fn = '\0';
    select = lstsel = -1;
    nf = 0;
    for (;;) {
	yy = iby1 + r * 2 - 4;
	while (select < 0 || select >= nf) {
	    for (fs = fn; *fs == ' '; fs++)
		continue;
	    if (fs > fn)
		strcpy(fn, fs);
	    g_bar(ibx1 + 2, yy - 10, ibx2 - 2, yy + r + 2, hsbgc);
	    g_settextclr(hsfgc, hsbgc);
	    getcwd(cd, PATHSZ);
	    sprintf(ds, "Current directory: %s", cd);
	    g_text(ibx1 + c, yy - 1, ds);
	    g_text(ibx1 + c, yy + r + 1, prmpt);
	    tc = g_getstr(fn, nc);
	    select = (tc == MDL || tc == '\t' || tc & FN) ? 0 : -1;
	    if (tc == MDL) {
		get_mouse_location(&mx, &my);
		cs = ((mx - ibx1) / c - x0) / cc;
		ys = ((my - iby1 + 8) / r);
		select = (ys - y0) + (hh - y0) * cs;
	    }
	    wc = (strchr(fn, '*') || strchr(fn, '?'));
	    if (wc)
		break;
	    if (change_dir(fn)) {
		break;
	    }
	    getcwd(ncd, PATHSZ);
	    if (strcmp(cd, ncd)) {
		*fn = '\0';
		g_bar(ibx1, iby1, ibx2, iby2, hsbgc);
		g_box(ibx1 + 1, iby1 + 1, ibx2 - 1, iby2 - 1, hsfgc);
	    }
	}
        lstsel = -1;
	strcpy(ds, fn);
	while (select >= 0 && select < nf) {
	    if (select == 0 && diroff > 0) {
		diroff = 0;
		break;
	    }
	    if (select != lstsel) {
		if (lstsel >= 0) {
		    strncpy(nm, filnam + (FL * lstsel), 15);
    		    nm[15] = '\0';
		    g_text(lxs, lys, nm);
		}
		xs = x0 + (select / rn) * cc;
		ys = y0 + select % rn;
		lxs = ibx1 + c * xs;
		lys = iby1 + r * ys;
		g_settextclr(hsbgc, hsfgc);
		strncpy(nm, filnam + (FL * select), 15);
		nm[15] = '\0';
		g_text(lxs, lys, nm);
		g_settextclr(hsfgc, hsbgc);
		lstsel = select;
	    }
	    tc = getevent();
	    if (tc == MDL) {
		get_mouse_location(&mx, &my);
		cs = ((mx - ibx1) / c - x0) / cc;
		ys = ((my - iby1 + 8) / r);
		select = (ys - y0) + (hh - y0) * cs;
	    } else if (tc == '\t' || tc == FK_Down_Arrow) {
		select++;
	    } else if (tc == FK_Shift_Tab || tc == FK_Up_Arrow) {
		select--;
	    } else if (tc == FK_Left_Arrow) {
		select -= rn;
	    } else if (tc == FK_Right_Arrow) {
		select += rn;
	    } else if (tc == '\r' || tc == MDR) {
		tc = '\r';
		break;
	    } else if (tc == ESC) {
		break;
	    }
	    if (select <= 0 && diroff >= rn) {
		select += rn;
		diroff -= rn;
		break;
	    }
	    if (select >= (nf - 1) && !done) {
		select -= rn;
		diroff += rn;
		break;
	    }
	}
	if (tc == ESC) {
            *fn = '\0';
            break;
	} else if (tc == '\r') {
	    if (!wc) {
	        break;
	    } else if (select >= 0 && select < nf) {
                for (fs = s = fn; *fs; fs++)
                    if (*fs == '/' || *fs == '\\')
                        s = fs + 1;
		strcpy(s, filnam + (FL * select));
		trim(s);
		if (change_dir(s))
                    break;
                select = lstsel = -1;
		tc = '\1';
		strcpy(fn, "*");
            }
	}
	while (chkevent())
	    (void) getevent();
	if (h == hh * r) {
	    g_bar(ibx1 + 2, yy + 1, ibx2 - 2, iby2 - 2, hsbgc);
	} else {
	    g_rstscr(ibss, ibx1, iby1);
	    h = hh * r;
	    iby1 = (scrszy - h) / 2;
	    iby2 = iby1 + h - 1;
	    ibss = g_savscr(ibx1, iby1, ibx2, iby2);
	    if (ibss < 0) {
		putmsgwk(errmsg);
		return;
	    } else {
		g_bar(ibx1, iby1, ibx2, iby2, hsbgc);
		g_box(ibx1 + 1, iby1 + 1, ibx2 - 1, iby2 - 1, hsfgc);
	    }
	}
	for (fs = fn; *fs == ' '; fs++)
	    continue;
#ifdef WIN32
	hnd = FindFirstFile(fs, &f);
	done = (hnd == INVALID_HANDLE_VALUE);
	for (nf = 0; nf < diroff && !done; nf++)
	    done = !FindNextFile(hnd, &f);
	for (nf = 0; nf < MAXNF && !done; nf++) {
	    sprintf(filnam + (FL * nf), "%-*.*s", FNMSZ, FNMSZ, (char *) f.cFileName);
	    done = !FindNextFile(hnd, &f);
	}
        FindClose(hnd);
#else /* WIN32 */
	done = _dos_findfirst(fs, _A_NORMAL, &f);
	for (nf = 0; nf < diroff && !done; nf++)
	    done = _dos_findnext(&f);
	for (nf = 0; nf < MAXNF && !done; nf++) {
	    sprintf(filnam + (FL * nf), "%-*.*s", FNMSZ, FNMSZ, f.name);
	    done = _dos_findnext(&f);
	}
#endif /* WIN32 */
	if (diroff > 0)
	    strcpy(filnam, " ... ");
	if (!done)
	    strcpy(filnam + FL * (MAXNF - 1), " ... ");
	x = x0;
	y = y0;
	g_settextclr(hsfgc, -1);
	for (i = 0; i < nf; i++, y++) {
	    if (y >= hh) {
		x += cc;
		y = y0;
	    }
	    strncpy(nm, filnam + (FL * i), 15);
	    nm[15] = '\0';
	    g_text(ibx1 + c * x, iby1 + r * y, nm);
	}
    }
    g_rstscr(ibss, ibx1, iby1);
    FREE(filnam);
    for (fs = fn; *fs == ' '; fs++)
	continue;
    if (fs > fn)
	strcpy(fn, fs);
    for (fs = fn + strlen(fn); *(fs - 1) == ' '; fs--)
	continue;
    *fs = '\0';
}
