/* g_gr.c */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <grlib.h>
#include "video.h"
#include "memory.h"
#include "fk.h"

#define MAXNSS 1
#define limit(mn,aa,mx) (((aa)<(mn))?(mn):((aa)>(mx))?(mx):(aa))

#ifndef SIGPTRDEF
#ifdef HUGE_PTR
typedef short huge *SIGPTR;
#else
typedef short *SIGPTR;
#endif /* HUGE_PTR */
#define SIGPTRDEF
#endif  /* SIGPTRDEF */

int     getevent();
void    beep();

extern  monoflg;

int     grfcrd;			/* indicates type of graphic card */
int     scrszx, scrszy;		/* screen size x and y */
int     boxx1, boxy1;		/* box lower-left corner */
int     boxx2, boxy2;		/* box upper-right corner */
int     cw1x1, cw1y1;		/* closeup window 1 lower-left  corner */
int     cw1x2, cw1y2;		/* closeup window 1 upper-right corner */
int     cw2x1, cw2y1;		/* closeup window 2 lower-left  corner */
int     cw2x2, cw2y2;		/* closeup window 2 upper-right corner */
int     hlpx1, hlpy1;		/* help screen lower-left  corner */
int     hlpx2, hlpy2;		/* help screen upper-right corner */
int     lvlx1, lvly1;		/* level window lower-left  corner */
int     lvlx2, lvly2;		/* level window upper-right corner */
int     wavcol; 		/* wave text color */
int     tlbcol, stscol;		/* toolbar and status text color */
int     boxcol, active;		/* nonactive and active box colors */
int     csrcol, scwcol;		/* cursor and scroll window colors */
int     mkfgc, scbgc;		/* marker foregound/screen background colors */
int     hsfgc, hsbgc;		/* help screen foreground/background colors */
int     msfgc, msbgc;		/* help screen foreground/background colors */
int     wnfgc, wnbgc;		/* window number foreground/background colors */
int     blfgc, blbgc;		/* box label foreground/background colors */
int     bmfgc, bmbgc;		/* box message foreground/background colors */
int     clfgc, clbgc;		/* closeup label foreground/background colors */
int     msgx, msgy;		/* message starting position */
int     stsx, stsy;		/* status starting position */
int     bxlx, bxly;		/* box label starting position */
int     bxmx, bxmy;		/* box message starting position */
int     cw1x, cw1y;		/* closeup window 1 label starting position */
int     cw2x, cw2y;		/* closeup window 2 label starting position */
int     cumx, cumy;		/* cursor message starting position */
int     ch_wd, ch_ht, ln_sp;	/* character width, height, and line spacing */

static void setcol();
static void setpos();

static int orig_graf_mode = 0;
static int sb_on = 0, sb_x1 = 0, sb_y1 = 0, sb_x2 = 0, sb_y2 = 0, sb_c = 0;
static int cu_on = 0, cu_h = 0, cu_x = 0, cu_y = 0, cu_c = 0;
static int nss = 0;
static unsigned char *scrnbuf[MAXNSS];
static int firstime = 1;
static int txt_bgc = 0;
static int txt_fgc = 0;

int
g_videotype()
{
    return (gr_adapt());
}

void
g_init()
{
    extern  _gr_xpix, _gr_ypix;
    static int grafmode = 0;

    if (firstime) {
	orig_graf_mode = gr_gtmod();
	switch (grfcrd) {
	case 0:		/* EGA */
	    grafmode = 16;
	    break;
	case 1:		/* HGC */
	    grafmode = 20;
	    break;
	case 2:		/* CGA */
	    grafmode = 6;
	    break;
	case 3:		/* VGA */
	    grafmode = 18;
	    break;
	case 4:		/* PEGA */
	    grafmode = 80;
	    break;
	default:
	    grafmode = 6;
	}
    }
    gr_init(grafmode, 0, 0);
    gr_clear(scbgc);
    gr_scrsiz(&scrszx, &scrszy);

    ch_wd = 8;
    ch_ht = 8;
    ln_sp = (scrszy > 250) ? 10 : 9;
    cu_on = 0;
    setcol();
    if (firstime) {
	setpos();
	firstime = 0;
    }
}

void
g_close()
{
    if (!firstime)
	gr_init(orig_graf_mode, 0, 0);
}

void
g_cursor_del()
{
    if (cu_on) {		/* remove the old line */
	gr_dotty(3);
	gr_line(cu_x, cu_y, cu_x, cu_y + cu_h, cu_c);
	gr_dotty(0);
	cu_on = 0;
    }
}

void
g_cursor_init(h, col)
{
    g_cursor_del();
    cu_h = h;
    cu_c = col;
}

void
g_cursor_move(x, y)
{
    g_cursor_del();
    gr_dotty(3);
    cu_x = x;
    cu_y = y;
    gr_line(cu_x, cu_y, cu_x, cu_y + cu_h, cu_c);
    gr_dotty(0);
    cu_on = 1;
}

void
g_sb_del()
{
    if (sb_on != 0) {
	gr_dotty(3);
	gr_line(sb_x1, sb_y1, sb_x2, sb_y1, sb_c);
	gr_line(sb_x2, sb_y1, sb_x2, sb_y2, sb_c);
	gr_line(sb_x2, sb_y2, sb_x1, sb_y2, sb_c);
	gr_line(sb_x1, sb_y2, sb_x1, sb_y1, sb_c);
	gr_dotty(0);
	sb_on = 0;
    }
}

void
g_sb_move(x1, y1, x2, y2, color)
{
    g_sb_del();
    gr_dotty(3);
    sb_x1 = x1;
    sb_y1 = y1;
    sb_x2 = x2;
    sb_y2 = y2;
    sb_c = color;
    gr_line(sb_x1, sb_y1, sb_x2, sb_y1, sb_c);
    gr_line(sb_x2, sb_y1, sb_x2, sb_y2, sb_c);
    gr_line(sb_x2, sb_y2, sb_x1, sb_y2, sb_c);
    gr_line(sb_x1, sb_y2, sb_x1, sb_y1, sb_c);
    gr_dotty(0);
    sb_on = 1;
}

void
g_settextclr(fg, bg)
{
    gr_settc(fg, bg);
    txt_fgc = fg;
    txt_bgc = bg;
}

void
g_text(x, y, s)
char   *s;
{
    gr_text(x, y, s);
}

void
g_gettp(px, py)
int *px ,*py;
{
    gr_gettp(px, py);
}

void
g_erase(int x, int y, int n)
{
    y += ln_sp - ch_ht;
    gr_rectf(x, y - ln_sp, x + ch_wd * n, y, txt_bgc);
}

void
g_txt_cur(int x, int y)
{
    y += ln_sp - ch_ht;
    gr_line(x, y, x + ch_wd, y, txt_fgc);
}

void
g_text_dif(x, y, s, t)
char   *s, *t;
{
    char c[2], temp;
    int i, ns, nt;

    ns = strlen(s);
    nt = strlen(t);    
    c[1] = '\0';
    for (i = 0; i < ns || i < nt; i++) {
        c[0] = (i < ns) ? s[i] : ' ';
        temp = (i < nt) ? t[i] : ' ';
        if (temp != c[0])
            gr_text(x + i * 8, y, c);
    }
}

void
g_bar(x1, y1, x2, y2, color)
{
    gr_rectf(x1, y1, x2, y2, color);
}

void
g_box(x1, y1, x2, y2, color)
{
    gr_recto(x1, y1, x2, y2, color);
}

void
g_line(x1, y1, x2, y2, color)
{
    gr_line(x1, y1, x2, y2, color);
}

void
g_setxor(flag)
{
    gr_dotty(flag ? 3 : 0);
}

void
g_move(x, y)
{
    gr_move(x, y);
}

void
g_draw(x, y, color)
{
    gr_draw(x, y, color);
}

/* g_drawmark - put one marker on the screen */

void
g_drawmark(x, y, c)
int     x, y, c;
{
    char    ch[2];
    int     x1, y1, x2, y2;

    if (x < 0 || x >= scrszx)
	return;
    if (y < 0 || y >= scrszy)
	return;
    if (c < 0 || c >= 26)
	return;
    g_setxor(1);
    x1 = x;
    y1 = y;
    x2 = x1;
    y2 = y + 8;
    gr_line(x1, y1, x2, y2, mkfgc);
    x1 -= 3;
    y1 = y + 18;
    ch[0] = 'A' + c;
    ch[1] = '\0';
    g_settextclr(mkfgc, 0);
    g_text(x1, y1, ch);
    g_setxor(0);
}

void
g_drawwave(pbuf, npts, nchn, xpix, ypix, xofst, yofst, yrange, yzero, color)
SIGPTR  pbuf;
int     xpix, ypix, xofst, yofst, color;
long    npts, yrange, yzero;
{
    int     x, y, lastx, ymn, ymx;
    int     yval, ymin, ymax, ytop, ybot;
    long    x0, x1, x2, y0, y1, y2, icount, i = 0;

    if (npts < 1 || nchn < 1 || yrange < 2)
	return;
    x1 = xpix;
    x2 = npts;
    y1 = ypix - 1;
    y2 = yrange - 1;
    ytop = yofst - ypix / 2;
    ybot = yofst + ypix / 2;

    x0 = xofst;
    y0 = yofst + (yzero * y1) / y2;

    x = (int) x0;
    yval = pbuf[i];
    i += nchn;
    y = (int) (y0 - (yval * y1) / y2);
    y = limit(ytop, y, ybot);
    gr_move(x, y);
    if (npts < 2) {
	gr_draw(x, y, color);
    } else if (x2 < x1 * 2) {
	for (icount = 1; icount < npts; icount++) {
	    x = (int) (x0 + (icount * x1) / x2);
	    yval = pbuf[i];
	    i += nchn;
	    y = (int) (y0 - (yval * y1) / y2);
	    y = limit(ytop, y, ybot);
	    gr_draw(x, y, color);
	}
    } else {
	for (icount = 1; icount < npts;) {
	    ymin = yval;
	    ymax = yval;
	    lastx = x;
	    while (x == lastx && icount < npts) {
		x = (int) (x0 + (icount++ * x1) / x2);
                yval = pbuf[i];
                i += nchn;
		if (yval < ymin)
		    ymin = yval;
		if (yval > ymax)
		    ymax = yval;
	    }
	    ymn = (int) (y0 - (ymin * y1) / y2);
	    ymx = (int) (y0 - (ymax * y1) / y2);
	    if (ymn > ybot)
		ymn = ybot;
	    if (ymx < ytop)
		ymx = ytop;
	    gr_rectf(x, ymx, x, ymn, color);
	}
    }
}

void
setcol()
{
    wavcol = 15;		/* wave color */
    stscol = 6; 		/* status text color */
    tlbcol = 7; 		/* toolbox text color */
    boxcol = 8;			/* nonactive box color */
    active = 5;			/* active box color */
    csrcol = 7;			/* cursor color */
    scwcol = 8;			/* scroll window color */
    mkfgc = 15;			/* marker color */
    scbgc = 0;			/* screen background color */
    msfgc = 14;			/* message forground color */
    msbgc = 0;			/* message background color */
    wnbgc = 1;			/* window background color */
    hsfgc = 15;			/* help screen foreground color */
    hsbgc = 6;			/* help screen background color */
    blfgc = 14;			/* box label foreground color */
    blbgc = 1;			/* box label background color */
    bmfgc = 14;			/* box message foreground color */
    bmbgc = 1;			/* box message background color */

    if (grfcrd == 1 || grfcrd == 2) {	/* fix colors for CGA & HGC */
	blfgc = bmfgc = msfgc = hsfgc = mkfgc = 7;
        stscol = tlbcol = 7;
	wavcol = boxcol = active = csrcol = scwcol = 1;
	blbgc = bmbgc = msbgc = hsbgc = scbgc = 0;
	wnbgc = scbgc = 0;
    } else if (monoflg) {
	blfgc = bmfgc = msfgc = hsfgc = 0;
        stscol = tlbcol = 0;
	wavcol = boxcol = active = scwcol = 0;
	blbgc = bmbgc = msbgc = hsbgc = scbgc = 15;
	wnbgc = scbgc = 15;
	csrcol = mkfgc = 15;
    } 
}

void
setpos()
{
    int     botscr;

    botscr = scrszy - 1;

    lvlx1 = 300;
    lvly1 = 20;
    lvlx2 = 400;
    lvly2 = botscr - 20;

    bxly = boxy1 - 2;		/* box label starting position, y */
    cumx = scrszx / 2;		/* cursor message starting position, x */
    cumy = bxly;		/* cursor message starting position, y */
    stsx = boxx1;		/* status starting position, x */
    stsy = botscr - 2;		/* status starting position, y */
    msgx = stsx;		/* message starting position, x */
    msgy = stsy - ln_sp;	/* message starting position, y */
}

void
beep()
{
    gr_beep();
}

int
g_savscr(x1, y1, x2, y2)
{
    long    nb;

    if (nss >= MAXNSS)
	return (-1);
    nb = gr_imagesize(x1, y1, x2, y2);
    if (nb >= 65000)
	return (-2);
    scrnbuf[nss] = (unsigned char *) MALLOC(nb, 1);
    if (scrnbuf[nss] == NULL)
	return (-3);
    gr_getimage(x1, y1, x2, y2, scrnbuf[nss]);
    return (nss++);
}

void
g_rstscr(ss, x1, y1)
{
    if (ss < 0 || ss >= nss || ss >= MAXNSS)
	return;
    gr_putimage(x1, y1, scrnbuf[ss], 0);
    FREE(scrnbuf[ss]);
    nss--;
}
