/* g_grx.c */

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <grx20.h>
#ifdef DJC
#include <bios.h>
#include <dos.h>
#define _REGS       REGS
#define _int86      int86
#endif /* DJC */
#include "waved.h"
#include "video.h"
#include "memory.h"
#include "fk.h"

#define MAXNSS 1
#define MXNPL 8
#define limit(mn,aa,mx) (((aa)<(mn))?(mn):((aa)>(mx))?(mx):(aa))

#ifndef SIGPTRDEF
typedef short *SIGPTR;
#define SIGPTRDEF
#endif  /* SIGPTRDEF */

void    beep();
void    mouse_shows();

extern  int monoflg;

int     grfcrd;			/* indicates type of graphic card */
int     scrszx, scrszy;		/* screen size x and y */
int     boxx1, boxy1;		/* box lower-left corner */
int     boxx2, boxy2;		/* box upper-right corner */
int     cw1x1, cw1y1;		/* closeup window 1 lower-left  corner */
int     cw1x2, cw1y2;		/* closeup window 1 upper-right corner */
int     cw2x1, cw2y1;		/* closeup window 2 lower-left  corner */
int     cw2x2, cw2y2;		/* closeup window 2 upper-right corner */
int     hlpx1, hlpy1;		/* help screen lower-left  corner */
int     hlpx2, hlpy2;		/* help screen upper-right corner */
int     lvlx1, lvly1;		/* level window lower-left  corner */
int     lvlx2, lvly2;		/* level window upper-right corner */
int     wavcol; 		/* wave text color */
int     tlbcol, stscol;		/* toolbar and status text color */
int     boxcol, active;		/* nonactive and active box colors */
int     csrcol, scwcol;		/* cursor and scroll window colors */
int     mkfgc, scbgc;		/* marker foregound/screen background colors */
int     hsfgc, hsbgc;		/* help screen foreground/background colors */
int     msfgc, msbgc;		/* help screen foreground/background colors */
int     wnfgc, wnbgc;		/* window number foreground/background colors */
int     blfgc, blbgc;		/* box label foreground/background colors */
int     bmfgc, bmbgc;		/* box message foreground/background colors */
int     clfgc, clbgc;		/* closeup label foreground/background colors */
int     msgx, msgy;		/* message starting position */
int     stsx, stsy;		/* status starting position */
int     bxlx, bxly;		/* box label starting position */
int     bxmx, bxmy;		/* box message starting position */
int     cw1x, cw1y;		/* closeup window 1 label starting position */
int     cw2x, cw2y;		/* closeup window 2 label starting position */
int     cumx, cumy;		/* cursor message starting position */
int     ch_wd, ch_ht, ln_sp;	/* character width, height, and line spacing */

static void setcol();
static void setpos();

static int firstime = 1;
static int sb_on = 0, sb_x1 = 0, sb_y1 = 0, sb_x2 = 0, sb_y2 = 0, sb_c = 0;
static int cu_on = 0, cu_h = 0, cu_x = 0, cu_y = 0, cu_c = 0;
static int nss = 0;
static unsigned char *scrnbuf[MAXNSS];
static int gr_cx = 0;
static int gr_cy = 0;
static int txt_x = 0;
static int txt_y = 0;
static int txtfgc = 0;
static int txtbgc = 0;

/************ lower-level mouse functions ********************************/

#define	MINT	0x33

/* returns 1 if mouse is present and resets mouse, otherwise returns 0 */
int
g_minit()
{
    union _REGS regs;

    regs.x.ax = 0;		/* function 0, reset and status */
    (void) _int86(MINT, &regs, &regs);
    return ((regs.x.ax == 0xFFFF) ? 1 : 0);
}

/* use nonzero flag to turn cursor on,  flag = 0 to turn cursor off */
void
g_mshow(int flag)
{
    union _REGS regs;

    regs.x.ax = (flag == 0) ? 2 : 1;	/* function 1: show cursor, funct 2:
					   hide */
    (void) _int86(MINT, &regs, &regs);
    return;
}

/* get current status of buttons and location, store in m */
static void
g_mstatus(int *but1, int *but2, int *x, int *y)
{
    union _REGS regs;

    regs.x.ax = 3;		/* function 3: get button status and mouse
				   position */
    (void) _int86(MINT, &regs, &regs);
    *but1 = (regs.x.bx & 1) == 1;	/* check bit 0 */
    *but2 = (regs.x.bx & 2) == 2;	/* check bit 1 */
    *x = regs.x.cx & 0xFFFF;
    *y = regs.x.dx & 0xFFFF;
    return;
}

/********************* grlib functions ***********************************/

void
gr_rectf(int x1, int y1, int x2, int y2, int c)
{
    GrFilledBox(x1, y1, x2, y2, c);
}

void
gr_recto(int x1, int y1, int x2, int y2, int c)
{
    GrBox(x1, y1, x2, y2, c);
}

void
gr_line(int x1, int y1, int x2, int y2, int c)
{
    GrLine(x1, y1, x2, y2, c);
}

void
gr_move(int x, int y)
{
    gr_cx = x;
    gr_cy = y;
}

void
gr_draw(int x, int y, int c)
{
    GrLine(gr_cx, gr_cy, x, y, c);
    gr_cx = x;
    gr_cy = y;
}

long
gr_imagesize(int x1, int y1, int x2, int y2)
{
    long s;
    int   w, h, o, p;

    w = (x2 > x1) ? (x2 - x1 + 1) : (x1 - x2 + 1);
    h = (y2 > y1) ? (y2 - y1 + 1) : (y1 - y2 + 1);
    o = sizeof(GrContext);
    p = GrPlaneSize(w, h);
    s = o + GrNumPlanes() * p;

    return (s);
}

void
gr_getimage(int x1, int y1, int x2, int y2, unsigned char *i)
{
    char *m[MXNPL];
    int   k, w, h, o, p;
    GrContext *c, *s;

    w = (x2 > x1) ? (x2 - x1 + 1) : (x1 - x2 + 1);
    h = (y2 > y1) ? (y2 - y1 + 1) : (y1 - y2 + 1);
    o = sizeof(GrContext);
    p = GrPlaneSize(w, h);
    m[0] = i + o;
    for (k = 1; k < GrNumPlanes(); k++)
        m[k] = m[k - 1] + p;
    GrCreateContext(w, h, m, (GrContext *) i);
    c = (GrContext *) i;
    s = (GrContext *) NULL;
    GrBitBltNC(c, 0, 0, s, x1, y1, x2, y2, GrWRITE);
}

void
gr_putimage(int x, int y, unsigned char *i, int m)
{
    int   xx, yy;
    GrContext *c, *s;

    c = (GrContext *) i;
    s = (GrContext *) NULL;
    xx = c->gc_xmax;
    yy = c->gc_ymax;
    GrBitBltNC(s, x, y, c, 0, 0, xx, yy, GrWRITE);
}

/********************************************************************/

int
g_videotype()
{
    return (0);
}

void
g_init()
{
    int     ncolors = 16;

    scrszx = 640;
    scrszy = 480;
    GrSetMode(GR_width_height_color_graphics, scrszx, scrszy, ncolors);
    signal(SIGINT, SIG_IGN);

    ch_wd = 8;
    ch_ht = 12;
    ln_sp = 16;
    cu_on = 0;
    setcol();
    if (firstime) {
	setpos();
	firstime = 0;
    }
}

void
g_close()
{
    GrSetMode(GR_default_text);
}

void
g_cursor_del()
{
    if (cu_on) {		/* remove the old line */
	gr_line(cu_x, cu_y, cu_x, cu_y + cu_h, cu_c | GrXOR);
	cu_on = 0;
    }
}

void
g_cursor_init(h, col)
{
    g_cursor_del();
    cu_h = h;
    cu_c = col;
}

void
g_cursor_move(x, y)
{
    g_cursor_del();
    cu_x = x;
    cu_y = y;
    gr_line(cu_x, cu_y, cu_x, cu_y + cu_h, cu_c | GrXOR);
    cu_on = 1;
}

void
g_sb_del()
{
    if (sb_on != 0) {
	gr_line(sb_x1, sb_y1, sb_x2, sb_y1, sb_c | GrXOR);
	gr_line(sb_x2, sb_y1, sb_x2, sb_y2, sb_c | GrXOR);
	gr_line(sb_x2, sb_y2, sb_x1, sb_y2, sb_c | GrXOR);
	gr_line(sb_x1, sb_y2, sb_x1, sb_y1, sb_c | GrXOR);
	sb_on = 0;
    }
}

void
g_sb_move(x1, y1, x2, y2, color)
{
    g_sb_del();
    sb_x1 = x1;
    sb_y1 = y1;
    sb_x2 = x2;
    sb_y2 = y2;
    sb_c = color;
    gr_line(sb_x1, sb_y1, sb_x2, sb_y1, sb_c | GrXOR);
    gr_line(sb_x2, sb_y1, sb_x2, sb_y2, sb_c | GrXOR);
    gr_line(sb_x2, sb_y2, sb_x1, sb_y2, sb_c | GrXOR);
    gr_line(sb_x1, sb_y2, sb_x1, sb_y1, sb_c | GrXOR);
    sb_on = 1;
}

void
g_settextclr(int fg, int bg)
{
    txtfgc = fg;
    txtbgc = bg;
}

void
g_text(int x, int y, char *s)
{
    GrTextXY(x, y - 12, s, txtfgc, txtbgc);
    txt_x = x + ch_wd * strlen(s);
    txt_y = y;
}

void
g_gettp(int *px, int *py)
{
   *px = txt_x;
   *py = txt_y;
}

void
g_erase(int x, int y, int n)
{
    y += ln_sp - ch_ht;
    gr_rectf(x, y - ln_sp, x + ch_wd * n, y, txtbgc);
}

void
g_txt_cur(int x, int y)
{
    y += ln_sp - ch_ht;
    gr_line(x, y, x + ch_wd, y, txtfgc);
}

void
g_text_dif(x, y, s, t)
char   *s, *t;
{
    char c[2], temp;
    int i, ns, nt;

    ns = strlen(s);
    nt = strlen(t);    
    c[1] = '\0';
    for (i = 0; i < ns || i < nt; i++) {
        c[0] = (i < ns) ? s[i] : ' ';
        temp = (i < nt) ? t[i] : ' ';
        if (temp != c[0])
            g_text(x + i * 8, y, c);
    }
}

void
g_bar(int x1, int y1, int x2, int y2, int color)
{
    gr_rectf(x1, y1, x2, y2, color);
}

void
g_box(int x1, int y1, int x2, int y2, int color)
{
    gr_recto(x1, y1, x2, y2, color);
}

void
g_line(int x1, int y1, int x2, int y2, int color)
{
    gr_line(x1, y1, x2, y2, color);
}

void
g_move(int x, int y)
{
    gr_move(x, y);
}

void
g_draw(int x, int y, int color)
{
    gr_draw(x, y, color);
}

/* g_drawmark - put one marker on the screen */

void
g_drawmark(int x, int y, int c)
{
    char    ch[2];
    int     x1, y1, x2, y2;

    if (x < 0 || x >= scrszx)
	return;
    if (y < 0 || y >= scrszy)
	return;
    if (c < 0 || c >= 26)
	return;
    x1 = x;
    y1 = y;
    x2 = x1;
    y2 = y + 8;
    gr_line(x1, y1, x2, y2, mkfgc | GrXOR);
    x1 -= 3;
    y1 = y + 18;
    ch[0] = 'A' + c;
    ch[1] = '\0';
    g_settextclr(mkfgc | GrXOR, GrXOR);
    g_text(x1, y1, ch);
}

void
g_drawwave(pbuf, npts, nchn, xpix, ypix, xofst, yofst, yrange, yzero, color)
SIGPTR  pbuf;
int     xpix, ypix, xofst, yofst, color;
long    npts, yrange, yzero;
{
    int     x, y, lastx, ymn, ymx;
    int     yval, ymin, ymax, ytop, ybot;
    long    x0, x1, x2, y0, y1, y2, icount, i = 0;

    if (npts < 1 || nchn < 1 || yrange < 2)
	return;
    x1 = xpix;
    x2 = npts;
    y1 = ypix - 1;
    y2 = yrange - 1;
    ytop = yofst - ypix / 2;
    ybot = yofst + ypix / 2;

    x0 = xofst;
    y0 = yofst + (yzero * y1) / y2;

    x = (int) x0;
    yval = pbuf[i];
    i += nchn;
    y = (int) (y0 - (yval * y1) / y2);
    y = limit(ytop, y, ybot);
    gr_move(x, y);
    if (npts < 2) {
	gr_draw(x, y, color);
    } else if (x2 < x1 * 2) {
	for (icount = 1; icount < npts; icount++) {
	    x = (int) (x0 + (icount * x1) / x2);
	    yval = pbuf[i];
	    i += nchn;
	    y = (int) (y0 - (yval * y1) / y2);
	    y = limit(ytop, y, ybot);
	    gr_draw(x, y, color);
	}
    } else {
	for (icount = 1; icount < npts;) {
	    ymin = yval;
	    ymax = yval;
	    lastx = x;
	    while (x == lastx && icount < npts) {
		x = (int) (x0 + (icount++ * x1) / x2);
                yval = pbuf[i];
                i += nchn;
		if (yval < ymin)
		    ymin = yval;
		if (yval > ymax)
		    ymax = yval;
	    }
	    ymn = (int) (y0 - (ymin * y1) / y2);
	    ymx = (int) (y0 - (ymax * y1) / y2);
	    if (ymn > ybot)
		ymn = ybot;
	    if (ymx < ytop)
		ymx = ytop;
	    gr_rectf(x, ymx, x, ymn, color);
	}
    }
}

void
setcol()
{
    wavcol = 15;		/* wave color */
    stscol = 6; 		/* status text color */
    tlbcol = 7; 		/* toolbox text color */
    boxcol = 8;			/* nonactive box color */
    active = 5;			/* active box color */
    csrcol = 7;			/* cursor color */
    scwcol = 8;			/* scroll window color */
    mkfgc = 15;			/* marker color */
    scbgc = 0;			/* screen background color */
    msfgc = 14;			/* message forground color */
    msbgc = 0;			/* message background color */
    wnbgc = 1;			/* window background color */
    hsfgc = 15;			/* help screen foreground color */
    hsbgc = 6;			/* help screen background color */
    blfgc = 14;			/* box label foreground color */
    blbgc = 1;			/* box label background color */
    bmfgc = 14;			/* box message foreground color */
    bmbgc = 1;			/* box message background color */

    if (grfcrd == 1 || grfcrd == 2) {	/* fix colors for CGA & HGC */
	blfgc = bmfgc = msfgc = hsfgc = mkfgc = 7;
        stscol = tlbcol = 7;
	wavcol = boxcol = active = csrcol = scwcol = 1;
	blbgc = bmbgc = msbgc = hsbgc = scbgc = 0;
	wnbgc = scbgc = 0;
    } else if (monoflg) {
	blfgc = bmfgc = msfgc = hsfgc = 0;
        stscol = tlbcol = 0;
	wavcol = boxcol = active = scwcol = 0;
	blbgc = bmbgc = msbgc = hsbgc = scbgc = 15;
	wnbgc = scbgc = 15;
	csrcol = mkfgc = 15;
    } 
}

void
setpos()
{
    int     botscr;

    botscr = scrszy - 1;

    lvlx1 = 300;
    lvly1 = 20;
    lvlx2 = 400;
    lvly2 = botscr - 20;

    bxly = boxy1 - 2;		/* box label starting position, y */
    cumx = scrszx / 2;		/* cursor message starting position, x */
    cumy = bxly;		/* cursor message starting position, y */
    stsx = boxx1;		/* status starting position, x */
    stsy = botscr - 2;		/* status starting position, y */
    msgx = stsx;		/* message starting position, x */
    msgy = stsy - ln_sp;	/* message starting position, y */
}

void
beep()
{
    fputs("\a", stderr);
}

int
g_savscr(x1, y1, x2, y2)
{
    long    nb = 0;

    if (nss >= MAXNSS)
	return (-1);
    nb = gr_imagesize(x1, y1, x2, y2);
    if (nb >= 307200)
	return (-2);
    scrnbuf[nss] = (unsigned char *) MALLOC(nb, 1);
    if (scrnbuf[nss] == NULL)
	return (-3);
    gr_getimage(x1, y1, x2, y2, scrnbuf[nss]);
    return (nss++);
}

void
g_rstscr(ss, x1, y1)
{
    if (ss < 0 || ss >= nss || ss >= MAXNSS)
	return;
    gr_putimage(x1, y1, scrnbuf[ss], 0);
    FREE(scrnbuf[ss]);
    nss--;
}

/***********************************************************************/

void
gr_xshft(int y1, int y2, int nb)
{
    int x1, x2, y, w, h, b;

    if (nb > 0) {
    	b = nb * MNBS;
        x1 = 0;
        x2 = b;
    } else if (nb < 0) {
    	b = -nb * MNBS;
        x1 = b;
        x2 = 0;
    } else {
        return;
    }
    w = scrszx - b;
    if (y2 > y1) {
        h = y2 - y1 + 1;
        y = y1;
    } else {
        h = y1 - y2 + 1;
        y = y2;
    }
    GrBitBltNC(NULL, x2, y, NULL, x1, y, x1 + w, y + h, GrWRITE);
}

/***********************************************************************/

#define Ctrl_C      3
#define MAXNSS      1

static int mouse_evnt = 0;
static int mcx = 0, mcy = 0;
static int  keyval = 0;

int 
chkevent()
{
    unsigned c;
    int     ls, rs;
    static int lls = 0, lrs = 0;
    static int mlx = 0, mly = 0;

    if (keyval == 0) {
	c = 0;
	if (_bios_keybrd(_NKEYBRD_READY)) {
	    c = _bios_keybrd(_NKEYBRD_READ);
	    if ((c & 0xFF) == 0x00 ||    /* old scan code */
	    	(c & 0xFF) == 0xE0) {   /* new scan code */
		c = FN | ((c >> 8) & 0xFF);
	    } else if ((c & 0xFF)) {
		c &= 0xFF;
	        if (c == Ctrl_C) {
	            keyval = c;
	        }
	    }
	}
	if (mouse_evnt) {
	    g_mstatus(&ls, &rs, &mcx, &mcy);
       	    if (!ls && lls) { /* check left button release */
		c = MDL;
	    } else if (!rs && lrs) { /* check right button release */
		c = MDR;
	    } else if (mcx != mlx || mcy != mly) {
	        mlx = mcx;
	        mly = mcy;
	        if (mouse_evnt > 1)
	            c = MMV;
		mouse_shows(1);
	    }
	    lls = ls;
            lrs = rs;
	}
	keyval = c;
    }
    return (keyval);
}

void
get_mouse_location(int *x, int *y)
{
    *x = mcx;
    *y = mcy;
}

int
getevent()
{
    int c;

    mouse_shows(1);
    while (keyval == 0)
	chkevent();
    c = keyval;
    keyval = 0;
    mouse_shows(0);
    return (c);
}

void
set_mouse_event(m)
{
    mouse_evnt = m;
}

int
lastevent()
{
    return (keyval);
}
