/* iod1.c - I/O Device = 8-bit DAC at I/O port (LPT, VMK, or other) */

#include <dos.h>
#include "iod.h"

#define todc	(*((long FAR *) 0x46C))
#define CTPS 	18.207
#define CPS     1193180L
#define TIMER   8

void stop_iod1();
int output_dac1(SIGPTR, long, long);
int input_adc1(SIGPTR, long, long);

int chkevent();
int getevent();

static void (INTERRUPT * oldtimer) ();
static void INTERRUPT dac_isr1();
static void INTERRUPT adc_isr1();
static SIGPTR isr_samp;
static long isr_count = 0;
static long isr_npts = 0;
static int isr_ticks = 0;
static int isr_port = 0;
static int first_io = 1;

int
init_iod1()
{
    adbits = 8;
    adzero = 128;
    if (iod_port < 0x100 || iod_port > 0x3FF)
	return (1);
    isr_port = iod_port;
    isr_count = 0;
    isr_npts = 0;
    if (lptflg)			/* disable printer */
	(void) _outp((unsigned int) iod_port + 2, 4);
    if (first_io) {
	oldtimer = _dos_getvect(TIMER);
	first_io = 0;
    } else if (_dos_getvect(TIMER) != oldtimer) {
	set_timer_rate(0);
	_dos_setvect(TIMER, oldtimer);
    }
    stop_iod = stop_iod1;
    output_dac = output_dac1;
    input_adc = input_adc1;
    return (0);
}

void
stop_iod1()
{
    if (isr_npts != 0) {
	isr_npts = 0;
	_dos_setvect(TIMER, oldtimer);
	set_timer_rate(0);
    }
    if (lptflg)			/* enable printer */
	(void) _outp((unsigned int) iod_port + 2, 12);
}

int
output_dac1(buf, npts, rate)
SIGPTR  buf;
long    npts, rate;
{
    if (npts <= 0)
	return (0);
    if (rate <= 20)
	return (1);
    isr_ticks = CPS / rate;
    isr_samp = buf;
    isr_count = npts;
    isr_npts = npts;
    _dos_setvect(TIMER, dac_isr1);
    set_timer_rate(isr_ticks);
    while (isr_count > 0 && chkevent() == 0)
	continue;
    todc += (((isr_npts - isr_count) * isr_ticks) >> 16) + 1;
    set_timer_rate(0);
    _dos_setvect(TIMER, oldtimer);
    isr_npts = 0;
    return (0);
}

int
input_adc1(buf, npts, rate)
SIGPTR  buf;
long    npts, rate;
{
    if (npts <= 0)
	return (0);
    if (rate <= 20)
	return (1);
    isr_ticks = CPS / rate;
    isr_samp = buf;
    isr_count = npts;
    isr_npts = npts;
    _dos_setvect(TIMER, adc_isr1);
    set_timer_rate(isr_ticks);
    while (isr_count > 0 && chkevent() == 0)
	continue;
    todc += (((isr_npts - isr_count) * isr_ticks) >> 16) + 1;
    set_timer_rate(0);
    _dos_setvect(TIMER, oldtimer);
    if (chkevent())
	(void) getevent();
    adnpts = isr_npts - isr_count;
    adrate = isr_ticks * CPS;
    isr_npts = 0;
    return (0);
}

static void INTERRUPT
dac_isr1()
{
    if (isr_count > 0) {
	isr_count--;
	(void) _outp((unsigned int) isr_port, (int) *isr_samp++);
    } else {
	set_timer_rate(0);
    }
    (void) _outp(0x20, 0x20);
}

static void INTERRUPT
adc_isr1()
{
    if (isr_count > 0) {
	isr_count--;
	*isr_samp++ = _inp((unsigned int) isr_port);
    } else {
	set_timer_rate(0);
    }
    (void) _outp(0x20, 0x20);
}
