/* loadfile.c */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <sys\types.h>
#include <io.h>
#include "waved.h"
#include "video.h"
#include "file.h"
#ifdef DJC
#include <unistd.h>
#endif

#ifdef _O_BINARY
#define O_MODE     _O_RDONLY | _O_BINARY
#else
#define O_MODE     O_RDONLY
#endif /* _O_BINARY */
#ifndef _open
#define _open       open
#define _read       read
#define _close      close
#define _lseek      lseek
#define _write      write
#define _swab       swab
#endif /* _open */

int     guess_file_type();
int     sigsize();
void    get_file_name();
void    window_new();
void    window_close();
void    exch_snd_hdr(SND_fmt *);

extern char bps[8];
extern int data_type;
extern int byte_swap;
extern int no_header;
extern long head_size;
extern long cust_head;
extern long cust_tail;

static int rbopen();
static int valid_data();
static long file_len(int);
static long newsize(int, long, int);
static void loadfile_trn();
static void loadfile_lst();
static void loadfile_ils();
static void loadfile_noh();
static void loadfile_mat();
static void loadfile_wav();
static void loadfile_snd();
static void loadfile_adf();
static void read_data(int, int, long, long, int, int, char *);
static void show_signal();

void
loadfile(fn, cstm)
char   *fn;
{
    int     cw = 0, lw = 0, ft;
    char    tfn[FNSZ], msg[80];

    if (sig[aw].siglen > 0) {
	if (*fn) {
	    if (guess_file_type(fn) < 0) {
		sprintf(msg, " Can't open '%s'.", fn);
		putmsgwk(msg);
		return;
	    }
	} else if (lsw < MAXNSW) {
	    cw = aw;
	    lw = lsw;
	    window_new();
	} else {
	    putmsg(" DELETE current contents of window? (y/N) ");
	    if (!getyes())
		return;
	    sigdel(aw);
	    showdata();
	}
    }
    if (*fn) {
	strcpy(tfn, fn);
	if (sig[aw].siglen > 0)
	    sigdel(aw);
    } else {
	strncpy(tfn, sig[aw].sigfnm, FNMSZ);
	if (*tfn < ' ' || *tfn > '~')
	    *tfn = '\0';
	get_file_name("Enter name of file: ", tfn, FNSZ);
    }
    ft = guess_file_type(tfn);
    file_type = (cstm && ft >= 0) ? FT_NOH : ft;
    switch (file_type) {
    case FT_ILS:
	loadfile_ils(tfn);
	break;
    case FT_TRN:
	loadfile_trn(tfn);
	break;
    case FT_LST:
	loadfile_lst(tfn);
	break;
    case FT_SND:
	loadfile_snd(tfn);
	break;
    case FT_NOH:
	loadfile_noh(tfn);
	break;
    case FT_WAV:
	loadfile_wav(tfn);
	break;
    case FT_MAT:
	loadfile_mat(tfn);
	break;
    case FT_ADF:
	loadfile_adf(tfn);
	break;
    default:
	if (lsw > lw && lsw == aw) {
	    window_close();
	    chaw(cw);
	}
	if (strlen(tfn) > 0) {
	    sprintf(msg, " Can't open '%s'.", tfn);
	    putmsgwk(msg);
	}
    }
}

static void
wswapb(short *a, int n)
{
    _swab((char *) a, (char *) a, n * 2);
}

static void
lswapb(long *a, int n)
{
    unsigned char *ch, tmp;

    while (n-- > 0) {
	ch = (unsigned char *) a++;
	tmp = ch[3];
	ch[3] = ch[0];
	ch[0] = tmp;
	tmp = ch[2];
	ch[2] = ch[1];
	ch[1] = tmp;
    }
}

static void
dswapb(double *a, int n)
{
    unsigned char *ch, tmp;

    while (n-- > 0) {
	ch = (unsigned char *) a++;
	tmp = ch[7];
	ch[7] = ch[0];
	ch[0] = tmp;
	tmp = ch[6];
	ch[6] = ch[1];
	ch[1] = tmp;
	tmp = ch[5];
	ch[5] = ch[2];
	ch[2] = tmp;
	tmp = ch[4];
	ch[4] = ch[3];
	ch[3] = tmp;
    }
}

int
guess_file_type(tfn)
char   *tfn;
{
    char   *h, *trn_hdr, type[4];
    int     i, n, file;
    long   *mat_hdr, *rif_hdr;
    short  *ils_hdr;
    ADF_fmt *adf_hdr;
    SND_fmt *snd;

    if (head_size >= 0) {
        cust_head = head_size;
        return (FT_NOH);
    } else if (no_header) {
        return (FT_NOH);
    }
    if ((file = _open(tfn, O_MODE)) == -1)
	return (-1);
    n = _read(file, (char *) buffer, sizeof(buffer));
    _close(file);
    if (n <= 0)
	return (-1);

    ils_hdr = (short *) buffer;	/* ILS */
    if (TYPE == SDF && NBLKS > 0 && SF > 0)
	return (FT_ILS);

    rif_hdr = (long *) buffer;	/* RIFF WAVE ? */
    if (rif_hdr[0] == 0x46464952 && rif_hdr[2] == 0x45564157)
	return (FT_WAV);

    adf_hdr = (ADF_fmt *) buffer;	/* DAT */
    if (strcmp((char *) adf_hdr, adf_key) == 0)
	return (FT_ADF);

    trn_hdr = (char *) buffer;	/* TRAN ? */
    if (strcmp(trn_hdr, trn_key) == 0 || strcmp(trn_hdr, "VMF") == 0)
	return (FT_TRN);

    snd = (SND_fmt *) buffer;	/* NeXT SND ? */
    if (snd->key == snd_key)
	return (FT_SND);
    lswapb(&snd->key, 1);
    if (snd->key == snd_key)
	return (FT_SND);
    lswapb(&snd->key, 1);

    mat_hdr = (long *) buffer;	                /* MATLAB ? */
    type[0] = (mat_hdr[0] / 1000) % 10;		/* M */
    type[1] = (mat_hdr[0] / 100) % 10;		/* O */
    type[2] = (mat_hdr[0] / 10) % 10;		/* P */
    type[3] = (mat_hdr[0] / 1) % 10;		/* T */
    if (type[3] == 0)
	if (type[0] == 0 || type[0] == 1)
	    if (type[1] == 0 || type[1] == 1)
                if (type[2] >= 0 && type[2] <= 4)
		    if (mat_hdr[3] == 0)
			if (mat_hdr[1] == 1 || mat_hdr[2] == 1)
			    if (mat_hdr[4] > 0)
				return (FT_MAT);

    h = (char *) buffer;	/* ASCII LIST ? */
    for (i = 0; i < n && h[i] > 0; i++)
	continue;
    if (i == n)
	return (FT_LST);

    return (FT_NOH);
}

static void
loadfile_lst(tfn)
char   *tfn;
{
    char    msg[80];
    float   v, maxmag, scl;
    int     f, nchn = 1;
    long    cnt, nsmp, rate;
    SIGPTR  sp;
    SIGNAL *s;
    FILE   *fp;

    fp = fopen(tfn, "r");
    if (fp == NULL) {
	sprintf(msg, " Can't open '%s'.", tfn);
	putmsgwk(msg);
	return;
    }
    s = &sig[aw];
    rate = def_rate;
    s->sgtime = time(NULL);
    strncpy(s->sigfnm, tfn, FNMSZ);
    blankmarks(aw);

    f = 0;
    nsmp = 0;
    maxmag = 0;
    while (fgets(msg, 80, fp))
	if (valid_data(msg, &v, &f)) {
	    if (maxmag < v)
		maxmag = v;
	    if (maxmag < -v)
		maxmag = -v;
	    nsmp++;
	}
    if (nsmp <= 0) {
	putmsgwk(" No signal in file. ");
	return;
    }
    if ((nsmp = newsize(aw, nsmp, nchn)) <= 0) {
	fclose(fp);
	return;
    }
    scl = (f && maxmag > 0) ? (float) pow(10.0, log10(32767.0 / maxmag)) : 1;

    sprintf(msg, " Loading (LST) file %s ... ", tfn);
    putmsg(msg);
    fseek(fp, 0L, 0);		/* rewind to start of file */
    sp = s->sigbuf;
    cnt = 0;
    while (fgets(msg, 80, fp) && cnt < nsmp)
	if (valid_data(msg, &v, &f)) {
	    sp[cnt] = (int) (f ? v * scl : v);
	    cnt++;
	} else if (strncmp("; rate=", msg, 7) == 0) {
            sscanf(msg + 7, "%ld", &rate);
	}
    s->siglen = cnt;
    s->sgnchn = nchn;
    s->adrate = s->darate = rate;
    s->sigfty = file_type;
    s->sigfmt = 0;
    s->sigdty = 0;
    s->sigbsw = 0;
    show_signal();
    fclose(fp);
}

static void
loadfile_trn(tfn)
char   *tfn;
{
    char    msg[80];
    int     status, file, nchn = 1;
    long    nsmp;
    int len;
    SIGNAL *s;
    TRNHDR *trn_hdr;
    static int format = DT_UNSIGNED_8;
    static long offset = TRNHDSZ;

    if ((file = rbopen(tfn)) < 0)
	return;

    s = &sig[aw];
    trn_hdr = (TRNHDR *) buffer;
    len = TRNHDSZ;
    status = _read(file, (char *) trn_hdr, len);
    if (status != len) {
	putmsgwk(" Error reading TRN file header. ");
	return;
    }
    s->adrate = trn_hdr->vadr;
    s->darate = trn_hdr->vdar;
    s->sgtime = trn_hdr->vstm;
    s->sigmin = trn_hdr->vsmn;
    s->sigmax = trn_hdr->vsmx;
    s->sgbits = BITS;
    s->sigzer = DACZ;

    nsmp = trn_hdr->vnum;
    sprintf(msg, " Loading (TRN) file %s ... ", tfn);
    read_data(file, format, offset, nsmp, nchn, 0, msg);
}

static void
loadfile_snd(tfn)
char   *tfn;
{
    char    msg[80];
    SND_fmt *sound;
    int     status, file, format, swap = 0, nchn;
    int len;
    long    srate, nsmp;
    long    offset;
    SIGNAL *s;

    if ((file = rbopen(tfn)) < 0)
	return;
    sound = (SND_fmt *) buffer;
    len = sizeof(SND_fmt);
    status = _read(file, (char *) sound, len);
    if (status != len) {
	putmsgwk(" Error reading SND file header. ");
	return;
    }
    if (sound->key != snd_key) {
        exch_snd_hdr(sound);		/* change the data format */
        if (sound->key != snd_key) {
            printf("SND magic number wrong: %ld!=%ld\n", sound->key, snd_key);
	    abort();
        }
        swap = 1;
    }
    srate = sound->rate;
    offset = sound->ofst;
    format = (int) sound->code;
    nsmp = sound->size / bps[format];
    nchn = sound->chan;

    s = &sig[aw];
    s->adrate = s->darate = srate;
    sprintf(msg, " Loading (SND) file %s ... ", tfn);
    read_data(file, format, offset, nsmp, nchn, swap, msg);
}

static void
loadfile_ils(tfn)
char   *tfn;
{
    char    msg[80];
    int     status, sr_exp, i, file;
    int     itok = 0, ntok = 0, swap = 0, nchn;
    long    toksiz;
    long    srate, nblk, nsmp;
    long    offset, blksiz = 512;
    short  *ils_hdr, *hitch;
    int len;
    SIGNAL *s, *t;
    static int format = DT_LINEAR_16;

    if ((file = rbopen(tfn)) < 0)
	return;
    ils_hdr = (short *) buffer;
    len = (unsigned int) blksiz;
    status = _read(file, (char *) ils_hdr, len);
    if (status != len) {
	putmsgwk(" Error reading ILS file header. ");
	return;
    }
    srate = SF;
    sr_exp = SE;
    nblk = NBLKS;
    nchn = NCHAN;

    s = &sig[aw];
    (void) _lseek(file, (long) ILSHDSZ, 0);
    status = _read(file, msg, len = 6);
    if (status == len && strcmp(msg, btni_key) == 0) {
	t = (SIGNAL *) buffer;
        (void) _lseek(file, FNMSZ + 5 * sizeof(short) + 4L, 1);
	status = _read(file, (char *) &(t->siglen), (4 + NMARK) * sizeof(long) + 8);
	s->adrate = t->adrate;
	s->darate = t->darate;
	s->sgtime = t->sgtime;
	s->sigfmt = t->sigfmt;
	for (i = 0; i < NMARK; i++)
	    s->sigmrk[i] = t->sigmrk[i];
    } else {
	if (srate > 0) {
	    if (sr_exp > 0 && sr_exp < 6)
		while (sr_exp-- > 0)
		    srate *= 10;
	    s->adrate = s->darate = srate;
	}
    }

    itok = 0;
    toksiz = 0;
    ntok = 0;
    if (INIT == 11111) {
	offset = (nblk + 1) * blksiz;
	(void) _lseek(file, offset, 0);
	hitch = (short *) buffer;
	len = blksiz;
	status = _read(file, (char *) hitch, len);
	if (status == len) {
	    ntok = hitch[0];
	    toksiz = hitch[1];
	}
    }
    if (ntok == 0) {
	offset = blksiz;
	nsmp = (file_len(file) - blksiz) / 2;
	sprintf(msg, " Loading (ILS) file %s ... ", tfn);
    } else if (ntok == 1) {
        offset = blksiz;
        nsmp = toksiz;
        sprintf(msg, " Loading token 1 from file: %s ... ", tfn);
        sprintf(sig[aw].sigfnm, "[token 1]");
    } else {
	sprintf(msg, " Enter token number (1-%d) or ALL: ", ntok);
	get_response(msg, msg, 20);
	itok = atoi(msg);
	if (msg[0] == 'a' || msg[0] == 'A') {
	    offset = blksiz;
	    nsmp = toksiz * ntok;
	    sprintf(msg, " Loading token file %s ... ", tfn);
            sprintf(sig[aw].sigfnm, "%d tokens", ntok);
	} else if (itok >= 1 && itok <= ntok) {
	    offset = blksiz + (itok - 1) * toksiz * sizeof(short);
	    nsmp = toksiz;
	    sprintf(msg, " Loading token %d from file: %s ... ", itok, tfn);
            sprintf(sig[aw].sigfnm, "token %d", itok);
	} else {
	    putmsgwk(" Invalid token number.");
	    _close(file);
	    return;
	}
    }

    read_data(file, format, offset, nsmp, nchn, swap, msg);
}

static void
loadfile_noh(tfn)
char   *tfn;
{
    char    msg[80];
    int     fd, nchn = 1;
    long    nsmp;

    if ((fd = rbopen(tfn)) < 0)
	return;
    nsmp = (file_len(fd) - cust_head - cust_tail) / bps[data_type];
    sprintf(msg, " Loading file %s ... ", tfn);
    read_data(fd, data_type, cust_head, nsmp, nchn, byte_swap, msg);
}

static void
loadfile_wav(tfn)
char   *tfn;
{
    char    msg[80], *chkid;
    int     fd, frmt, swap, status, nchn;
    long    nsmp, hdsz;
    long    rifsz, fmtsz, datsz, facsz;
    WAV_fmt wfmt;

    if ((fd = rbopen(tfn)) < 0)
	return;
    chkid = (char *) buffer;
    chkid[4] = '\0';
    status = _read(fd, chkid, 4);
    if (status != 4) {
	putmsgwk(" Error reading WAV file header. ");
	return;
    }
    if (strncmp(chkid, "RIFF", 4) != 0)
	return;
    status = _read(fd, (char *) &rifsz, 4);
    status = _read(fd, chkid, 4);
    if (strncmp(chkid, "WAVE", 4) != 0)
	return;
    status = _read(fd, chkid, 4);
    if (strncmp(chkid, "fmt ", 4) != 0)
	return;
    status = _read(fd, (char *) &fmtsz, 4);
    status = _read(fd, (char *) &wfmt, (unsigned) fmtsz);
    hdsz = 7 * 4 + fmtsz;
    status = _read(fd, chkid, 4);
    while (strncmp(chkid, "fact", 4) == 0) {
        status = _read(fd, (char *) &facsz, 4);
        (void) _lseek(fd, facsz, 1);
        status = _read(fd, chkid, 4);
        hdsz += 2 * 4 + facsz;
    }
    if (strncmp(chkid, "data", 4) != 0)
	return;
    status = _read(fd, (char *) &datsz, 4);
    swap = 0;
    nchn = limit(1, wfmt.channels, MAX_CHAN);
    if (wfmt.bits_smp == 24) {
        frmt = DT_LINEAR_24;
        nsmp = datsz / 3 / nchn;
    } else if (wfmt.bits_smp == 16) {
        frmt = DT_LINEAR_16;
        nsmp = datsz / 2 / nchn;
    } else {
        frmt = DT_UNSIGNED_8;
        nsmp = datsz / nchn;
    }
    sig[aw].adrate = wfmt.samp_sec;
    sig[aw].darate = sig[aw].adrate;
    sprintf(msg, " Loading file %s ... ", tfn);
    read_data(fd, frmt, hdsz, nsmp, nchn, swap, msg);
}

static double
rd_mat_val(int fd, int dtyp, int bswp)
{
    double dtmp;
    float  ftmp;
    long   ltmp;
    short  stmp;
    unsigned long utmp;
    
    if (dtyp == 0) {
        (void) _read(fd, &dtmp, 8);
        if (bswp) dswapb(&dtmp, 1);
    } else if (dtyp == 1) {
        (void) _read(fd, &ftmp, 4);
        if (bswp) lswapb((long *) &ftmp, 1);
        dtmp = ftmp;
    } else if (dtyp == 2) {
        (void) _read(fd, &ltmp, 4);
        if (bswp) lswapb(&ltmp, 1);
        dtmp = ltmp;
    } else if (dtyp == 3) {
        (void) _read(fd, &stmp, 2);
        if (bswp) wswapb((short *) &stmp, 1);
        dtmp = stmp;
    } else if (dtyp == 4) {
        (void) _read(fd, &utmp, 4);
        if (bswp) lswapb((long *) &utmp, 1);
        dtmp = utmp;
    }
    return (dtmp);
}

static void
loadfile_mat(tfn)
char   *tfn;
{
    char    msg[80], nam[FNSZ];
    int     file, status, bsiz, dtyp, bswp = 0, nchn, type[4];
    long    hdr[5], ofst, nsmp, rate = 0;
    int len;
    static short bsz[5] = {8, 4, 4, 2, 4};
    static short dtp[5] = {
    	DT_DOUBLE, DT_FLOAT, DT_LINEAR_32, 
    	DT_LINEAR_16, DT_LINEAR_32
    };
    	

    if ((file = rbopen(tfn)) < 0)
	return;
    ofst = 0;
    for (;;) {
	(void) _lseek(file, ofst, 0);
	len = 20;
	status = _read(file, (char *) hdr, len);
	if (status != len) {
	    putmsgwk(" Error reading MAT file header. ");
	    return;
	}
        if (hdr[0] < 0 || hdr[0] > 9999 ||
            ((hdr[0] / 1000) % 10) != cpu_lsb) {
            bswp = 1;
        }
        if (bswp)
            lswapb(hdr, 5);
        type[0] = (hdr[0] / 1000) % 10;		/* M */
        type[1] = (hdr[0] / 100) % 10;		/* O */
        type[2] = (hdr[0] / 10) % 10;		/* P */
        type[3] = (hdr[0] / 1) % 10;		/* T */
	bswp = type[0];
	bsiz = bsz[type[2]];
	dtyp = dtp[type[2]];
        len = (unsigned int) hdr[4];
        status = _read(file, nam, len);
	ofst += 20 + hdr[4];
        if (strcmp(nam, "fmax") == 0) {
            rate = (long) rd_mat_val(file, type[2], bswp) * 2;
        } else if (strcmp(nam, "rate") == 0) {
            rate = (long) rd_mat_val(file, type[2], bswp);
        }
        if (hdr[1] > 1 && hdr[1] < MAX_CHAN) {
            nchn = hdr[1];
            nsmp = hdr[2];
        } else {
            nchn = 1;
            nsmp = hdr[1] * hdr[2];
        }
	if (nsmp > 1 && type[3] == 0 && hdr[3] == 0)
	    break;
	ofst += nsmp * bsiz;
	if (hdr[3])                 /* complex data */
            ofst += nsmp * bsiz;
    }
    if (nsmp <= 0)
        return;
    sig[aw].adrate = rate;
    sig[aw].darate = sig[aw].adrate;
    sprintf(msg, " Loading %s from file %s ... ", nam, tfn);
    read_data(file, dtyp, ofst, nsmp, nchn, bswp, msg);
}

static void
loadfile_adf(tfn)
char   *tfn;
{
    char    msg[80];
    ADF_fmt *adf_hdr;
    int     status, file, nchn;
    long    srate, nsmp;
    int len;
    static int format = DT_LINEAR_16;
    static long hs = (int) sizeof(ADF_fmt);

    if ((file = rbopen(tfn)) < 0)
	return;
    adf_hdr = (ADF_fmt *) buffer;
    len = hs;
    status = _read(file, (char *) adf_hdr, len);
    if (status != len) {
	putmsgwk(" Error reading ADF file header. ");
	return;
    }
    srate = (long) (adf_hdr->rate * 1000 + 0.5);
    nsmp = adf_hdr->npts;
    nchn = adf_hdr->chan;
    sig[aw].adrate = sig[aw].darate = srate;
    sprintf(msg, " Loading (ADF) file %s ... ", tfn);
    read_data(file, format, hs, nsmp, nchn, 0, msg);
}

static long
newsize(num, size, nchn)
int     num, nchn;
long    size;
{
    char    msg[80];

    if (size <= 0) {
	putmsg(" No signal.");
	return (0);
    }
    if (size > MAX_SAM) {
	sprintf(msg, " Truncating signal from %ld to %ld samples.",
	    size, MAX_SAM);
	putmsgwk(msg);
	size = MAX_SAM;
    }
    if (sigsize(num, size, nchn) == 0)
	return (0);
    return (size);
}

static long
file_len(int fd)
{
    long curr, size, zero = 0;

    curr = _lseek(fd, zero, SEEK_CUR);  /* save the current offset */
    size = _lseek(fd, zero, SEEK_END);  /* size = offset of EOF */
    curr = _lseek(fd, curr, SEEK_SET); 	/* return to current offset */

    return (size);
}

static int
rbopen(fn)
char   *fn;
{
    char    msg[80];
    int     file;

    file = _open(fn, O_MODE);
    if (file == -1) {
	sprintf(msg, " Can't open '%s'.", fn);
	putmsgwk(msg);
    }
    sig[aw].adrate = sig[aw].darate = def_rate;
    sig[aw].sgtime = time(NULL);
    sig[aw].sigfmt = 0;
    strncpy(sig[aw].sigfnm, fn, FNMSZ);
    blankmarks(aw);
    return (file);
}

double
scale_in(file, format, ip, np, swap, nsmp)
long    ip, np, nsmp;
{
    char   *bp;
    double *dbuf, v, vmx, sc;
    float  *fbuf;
    int     status, ns;
    long    k, j;
    int len;

    vmx = 0;
    (void) _lseek(file, ip, 0);
    bp = (char *) buffer;
    status = len = sizeof(buffer);
    for (k = 0, j = 0; k < nsmp && status == len; k += j) {
	status = _read(file, bp, len);
	switch (format) {
	case DT_FLOAT:
	    fbuf = (float *) bp;
            ns = status / 4;
            if (swap)
                lswapb((long *) fbuf, ns);
	    for (j = 0; (j < ns) && (k + j) < nsmp; j++) {
		v = fbuf[j];
		if (vmx < v)
		    vmx = v;
		if (vmx < -v)
		    vmx = -v;
	    }
	    break;
	case DT_DOUBLE:
	    dbuf = (double *) bp;
            ns = status / 8;
            if (swap)
                dswapb(dbuf, ns);
	    for (j = 0; (j < ns) && (k + j) < nsmp; j++) {
		v = dbuf[j];
		if (vmx < v)
		    vmx = v;
		if (vmx < -v)
		    vmx = -v;
	    }
	    break;
	}
    }
    sc = (vmx > 0) ? 32767.0 / vmx : 1;
    return (sc);
}

static void
read_data(int file, int format, long ip, long np, int nc, int swap, char *msg)
{
    char   *bp;
    double *dbuf, sc = 0;
    float  *fbuf;
    int     status, ns, nchn;
    long    nsmp, k, j, l, *lbuf;
    unsigned int len;
    SIGPTR  sp;
    SIGNAL *s;

    s = &sig[aw];
    nchn = limit(1, nc, MAX_CHAN);
    if ((nsmp = newsize(aw, np, nchn) * nchn) <= 0) {
	s->siglen = 0;
        s->sgnchn = 0;
    } else {
	putmsg(msg);
	if (format == DT_FLOAT || format == DT_DOUBLE)
	    sc = scale_in(file, format, ip, np, swap, nsmp);
	(void) _lseek(file, ip, 0);
	sp = s->sigbuf;
	bp = (char *) buffer;
	len = sizeof(buffer);
	if (format == DT_LINEAR_24)
	    len -= len % 3;
	for (k = 0, j = 0; k < nsmp; k += j) {
	    status = _read(file, bp, len);
	    if (status <= 0)
	    	break;
	    switch (format) {
	    case DT_UNSIGNED_8:
		for (j = 0; j < status && (k + j) < nsmp; j++)
		    sp[k + j] = (int) bp[j] & 255;
		break;
	    case DT_MULAW_8:
		for (j = 0; j < status && (k + j) < nsmp; j++)
		    sp[k + j] = ulaw2linear(bp[j]);
		break;
	    case DT_LINEAR_8:
		for (j = 0; j < status && (k + j) < nsmp; j++)
		    sp[k + j] = (int) bp[j];
		break;
	    case DT_LINEAR_16:
		if (swap)
		    _swab((char *) buffer, (char *) buffer, status);
                ns = status / 2;
		for (j = 0; j < ns && (k + j) < nsmp; j++)
		    sp[k + j] = (short) buffer[j];
		break;
	    case DT_LINEAR_24:
                ns = status / 3;
		for (j = 0; j < ns && (k + j) < nsmp; j++) {
		    l = 3 * j;
		    sp[k + j] = (int) ((bp[l + 1] & 0xFF)
			| ((bp[l + 2] & 0xFF) << 8));
		}
		break;
	    case DT_LINEAR_32:
		lbuf = (long *) bp;
		for (j = 0; j < (status / 4) && (k + j) < nsmp; j++)
		    sp[k + j] = (short) lbuf[j];
		break;
	    case DT_FLOAT:
		fbuf = (float *) bp;
                ns = status / 4;
                if (swap)
                    lswapb((long *) fbuf, ns);
                for (j = 0; (j < ns) && (k + j) < nsmp; j++)
		    sp[k + j] = (short) (fbuf[j] * sc);
		break;
	    case DT_DOUBLE:
		dbuf = (double *) bp;
                ns = status / 8;
                if (swap)
                    dswapb(dbuf, ns);
                for (j = 0; (j < ns) && (k + j) < nsmp; j++)
		    sp[k + j] = (short) (dbuf[j] * sc);
		break;
	    }
	}
	s->siglen = k / nchn;
        s->sgnchn = nchn;
	s->sigfty = file_type;
	s->sigdty = format;
	s->sigbsw = swap;
	show_signal();
    }
    _close(file);
}

static void
show_signal()
{
    int i, nc;
    long ns;
    
    ns = sig[aw].siglen;
    nc = limit(1, sig[aw].sgnchn, MAX_CHAN);
    scan_signal(aw);
    for (i = 0; i < nc; i++)
        sig[aw].sigbuf[ns * nc + i] = sig[aw].sigzer;
    curspnt = (ns < scrszx) ? ns / 2 : scrszx / 2;
    win[aw].winmrk1 = 0;
    win[aw].winmrk2 = ns;
    showdata();
    sigmsg(aw);
}

static int
valid_data(s, v, f)
char   *s;
float  *v;
int    *f;
{
    while (*s == ' ' || *s == '\t')
	s++;
    if (*s < '+' || *s > '9')
	return (0);
    if (strchr(s, ',') != NULL) {
	s = strchr(s, ',') + 1;
	while (*s == ' ' || *s == '\t')
	    s++;
	if (*s < '+' || *s > '9')
	    return (0);
    }
    if (strchr(s, '.') != NULL) {
	*v = (float) atof(s);
	if (*v != 0)
	    *f = 1;
    } else {
	*v = (float) atoi(s);
    }
    return (1);
}
