/*    QDRIVERS device driver package for Turbo C */

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>
#ifndef TURBOC
#include <malloc.h>
#else
#include <alloc.h>
#endif
#include "qdrive2.h"


char    Qdeftype;
char    Qcomptype;
char    Qgraphics;
unsigned Qbufsize;



unsigned Qbufdev[255];
char    Qbuftyp[255];
long    Qbufcnt[255];
qstring Qbufnme[255];
char    Qbufpag[255];
unsigned Qbase[15];
char    Qhard[15];
char    Qsize[15];
unsigned Qdev[15];
char    Qrammap[15][35];
char    Qpars[20];
char    Qcmd[15];
unsigned QDA1adda[15][35];
unsigned QDA1addb[15][35];
unsigned QDA1subs[15][5];
char    QDA1cmd[15];
unsigned QDA1size[15];
unsigned QDA1numsubs[15];
char    QDA1repa[15];
char    QDA1repb[15];
unsigned Qchans;
char    Qoldtype;
char    qdflag;
int     qelem;
int     qlaster;




/*******    TRASH    *********/



void 
Qerror(char qemess[], char qpro[])
{
    printf("** Error in procedure: %s \n", qpro);
    printf("     %s. \n\n", qemess);
    exit(0);
}

unsigned 
Qpointer(int qdevice, char qcaller[])
{

    unsigned qpnt;


    qpnt = (qdevice >> 7) + (qdevice & 0x01);
    if (qpnt < 2 || qpnt > 9 || Qbase[qpnt] == 0)
	Qerror("Illegal or non-present device specified", qcaller);
    return (qpnt);
}




/*******    DATA CONVERSION    *********/



unsigned 
Qactive(int qdevice)
{

    unsigned qpnt, qi;

    qi = 0;
    qpnt = Qpointer(qdevice, "Qactive");
    if (qpnt < 7) {
	qi = inportb(Qbase[qpnt] + QSTATUS) & QACT;
	qi = qi << 7;
    } else {
	qi = (inportb(Qbase[qpnt]) & 0x03) ^ 0x03;
    }
    return (qi);
}



void 
Qgo(int qdevice)
{
    unsigned qpnt;
    unsigned ql, qh, qi;


    if (qdevice == QALL) {
	ql = 2;
	qh = 8;
    } else {
	ql = Qpointer(qdevice, "QADgo");
	qh = ql;
    }
    for (qi = ql; qi <= qh; qi++) {
	if (Qdev[qi] > 0) {
	    qdevice = Qdev[qi];
	    qpnt = qi;
	    if (qdevice < QDA1) {
		outportb(Qbase[qpnt] + QTRIGGER, 0);
	    }
	    if ((qdevice & 0xFFFE) == QDA1) {
		outportb(Qbase[qpnt] + QDA1_CMD, QDA1cmd[qpnt]);
		outportb(Qbase[qpnt] + QDA1_SINK, 0);
		QDA1cmd[qpnt] = QDA1cmd[qpnt] && (QDA1_MANA || QDA1_MANB);
	    }
	}
    }
}


void 
Qsetcnt(int qdevice, long qcnt, char qcaller[])
{
    unsigned qpnt;
    long    qi;

    qpnt = Qpointer(qdevice, qcaller);
    if (qcnt < 0)
	Qerror("Illegal COUNT value specified", qcaller);
    qi = 0xFF0000 - qcnt;
    outport(Qbase[qpnt] + QCOUNT, qi & 0xFFFF);
    outportb(Qbase[qpnt] + QCOUNT + 2, qi >> 16);
    outportb(Qbase[qpnt] + QCOUNT + 2, qi >> 16);
}


void 
Qsrate(int qdevice, float qperiod, char qcaller[])
{
    unsigned qpnt;


    qpnt = Qpointer(qdevice, qcaller);
    if (qperiod < 0.05 || qperiod > 6000)
	Qerror("Illegal clock rate specified", qcaller);

    outport(Qbase[qpnt] + QCLOCK, (int) (65282.0 - qperiod * 10.0));
    outport(Qbase[qpnt] + QCLOCK, (int) (65282.0 - qperiod * 10.0));
}



void 
Qconvert(long qbuf, long qpnts, float qrate)
{
    unsigned qpnt, qb;
    char    qjunk;
    unsigned qdevice;
    int     qi;
    unsigned qd;

    qd = QDA_A;
    if (qbuf > 0x1000) {
	if (qbuf > 0x3000)
	    qd = QDA_A + QDA_B;
	else if (qbuf > 0x2000)
	    qd = QDA_B;
	qbuf = qbuf & 0xFF;
    }
    qpnt = Qbufdev[qbuf];
    qb = Qbase[qpnt];
    if (qpnt == 0)
	Qerror("Destination buffer not defined", "Qconvert");
    if (qpnt < 8) {
	qdevice = Qdev[qpnt];
	outportb(qb + QCOMMAND, 0x00);
	outportb(qb + QTRIGGER, 0);

	if (qdevice == QDA || qdevice == QDAx)
	    Qcmd[qpnt] = Qcmd[qpnt] & 0xBB;

	if (qpnts == QALL)
	    qpnts = Qbufcnt[qbuf];
	else if (qdevice == QAD || qdevice == QADx) {
	    outportb(qb + QRSTOVER, 0);
	    qpnts = qpnts * Qchans;
	}
	Qsetcnt(qdevice, qpnts, "Qconvert");
	Qsrate(qdevice, qrate, "Qconvert");
	outportb(qb + QPAGE, Qbufpag[qbuf] | QRAGRUN);
	outportb(qb + QPAGE, Qbufpag[qbuf] | QRAGRUN);
	outportb(qb + QPDDWSEL, 0);
	outportb(qb + QPDDWSEL, 0);
	if (qdevice == QDA || qdevice == QDAx)
	    Qcmd[qpnt] = Qcmd[qpnt] | qd;
	outportb(qb + QCOMMAND, Qcmd[qpnt]);
	outportb(qb + QCOMMAND, Qcmd[qpnt] | 0x80);
    }
    if (qpnt > 7) {
	outport(qb + QDA1_CLK, (int) (1.0 * 0x10000 - qrate * 20.0));
	if (qpnts != QALL)
	    Qerror("All buffer points must be specified on the QDA1, use QALL", "Qconvert");
	if (qbuf >= QCHAN_B) {
	    outportb(Qbase[qpnt] + QDA1_REPB, QDA1repb[qpnt]);
	    outport(qb + QDA1_ADDB, QDA1addb[qpnt][Qbufpag[qbuf]] | QDA1subs[qpnt][2]);
	    QDA1cmd[qpnt] = QDA1cmd[qpnt] | QDA1_GOB;
	} else {
	    outportb(Qbase[qpnt] + QDA1_REPA, QDA1repa[qpnt]);
	    outport(qb + QDA1_ADDA, QDA1adda[qpnt][Qbufpag[qbuf]] | QDA1subs[qpnt][1]);
	    QDA1cmd[qpnt] = QDA1cmd[qpnt] | QDA1_GOA;
	}
    }
}


void 
Qstop(int qdevice, int qmode)
{
    unsigned qpnt, qh, ql, qi;


    if (qdevice == QALL) {
	ql = 2;
	qh = 8;
	qmode = QDUAL;
    } else {
	ql = Qpointer(qdevice, "QADgo");
	qh = ql;
    }
    for (qi = ql; qi <= qh; qi++) {
	if (Qdev[qi] > 0) {
	    qdevice = Qdev[qi];
	    qpnt = qi;
	    if (qpnt < 8)
		if (qdevice == QAD || qdevice == QADx)
		    outportb(Qbase[qpnt] + QCOMMAND, Qcmd[qpnt] & 0x7F);
	    if (qdevice == QDA || qdevice == QDAx) {
		if (qmode == QCHAN_A || qmode == QCHAN_B)
		    Qerror("Individual channel control not supported on the QDA2", "Qstop");
		if (qmode == QDELAYED)
		    outportb(Qbase[qpnt] + QCOMMAND, Qcmd[qpnt] & 0x7F);
		else
		    outportb(Qbase[qpnt] + QCOMMAND, Qcmd[qpnt] & 0x77);
	    }
	    if (qpnt > 7) {
		if (qmode == QDELAYED)
		    Qerror("Delayed stopping not supported on the QDA1", "Qstop");
		else {
		    if (qmode == QCHAN_A)
			outportb(Qbase[qpnt] + QDA1_CMD, QDA1_STPA);
		    if (qmode == QCHAN_B)
			outportb(Qbase[qpnt] + QDA1_CMD, QDA1_STPB);
		    if (qmode == QDUAL)
			outportb(Qbase[qpnt] + QDA1_CMD, QDA1_STPA + QDA1_STPB);
		    outportb(Qbase[qpnt] + QDA1_SINK, 0);
		}
	    }
	}
    }
}


void 
Qtrig(int qdevice, int qtmode)
{
    unsigned qpnt, qi;


    qpnt = Qpointer(qdevice, "QADtrig");
    if (qpnt > 7)
	Qerror("No trigger modes supported.  The QDA1 defaults to MANUAL", "Qtrig");
    if (qtmode < 0 || qtmode > 15)
	Qerror("Illegal trigger mode specified", "QADtrig");
    Qcmd[qpnt] = (Qcmd[qpnt] & 0xF0) | qtmode;
}




/*******    QAD SPECIFIC PROCEDURES     *********/


void 
QADchan(int qdevice, int qpddw, int qchan)
{
    unsigned qpnt;
    unsigned ql, qh, qi;


    if (qpddw == QALL) {
	ql = 1;
	qh = 16;
	qpddw = 1;
    } else {
	ql = qpddw;
	qh = qpddw;
    }
    if (qpddw < 1 || qpddw > 16)
	Qerror("Illegal PDDW specified", "QADchan");
    if (qchan < 1 || qchan > 16)
	Qerror("Illegal analog input channel specified", "QADchan");
    qpnt = Qpointer(qdevice, "QADchan");
    for (qpddw = ql; qpddw <= qh; qpddw++) {
	outportb(Qbase[qpnt] + QPDDWSEL, qpddw - 1);
	outportb(Qbase[qpnt] + QAINSEL, (qchan - 1) ^ 0x0F);
    }
}


void 
QADcompare(int qdevice, int qpddw, int qflag)
{
    unsigned qpnt;
    unsigned qi, ql, qh;


    if (qpddw == QALL) {
	ql = 1;
	qh = qlaster - 1;
	qpddw = 1;
    } else {
	ql = qpddw;
	qh = qpddw;
    }
    qpnt = Qpointer(qdevice, "QADcompare");

    if (qpddw < 1 || qpddw > 16)
	Qerror("Illegal PDDW specified", "QADcompare");
    if (qflag < 0 || qflag > 1)
	Qerror("Illegal compare flag specified", "QADcompare");

    for (qi = ql; qi <= qh; qi++) {
	qpddw = qi;
	if (qlaster != 1)
	    qpddw = qpddw + 1;
	Qpars[qpddw] = (Qpars[qpddw] & 0x07) | (qflag << 3);
	outportb(Qbase[qpnt] + QPDDWSEL, qpddw - 1);
	outportb(Qbase[qpnt] + QAINPAR, Qpars[qpddw]);
    }
}


void 
QADgain(int qdevice, int qpddw, int qgain)
{
    unsigned qi, ql, qh;
    unsigned qpnt;

    if (qpddw == QALL) {
	ql = 1;
	qh = 16;
	qpddw = 1;
    } else {
	ql = qpddw;
	qh = qpddw;
    }
    qpnt = Qpointer(qdevice, "QADgain");

    if (qpddw < 1 || qpddw > 16)
	Qerror("Illegal PDDW specified", "QADgain");
    if (qgain < 0 || qgain > 3)
	Qerror("Illegal input GAIN specified", "QADgain");
    for (qpddw = ql; qpddw <= qh; qpddw++) {
	Qpars[qpddw] = (Qpars[qpddw] & 0x0C) | qgain;
	outportb(Qbase[qpnt] + QPDDWSEL, qpddw - 1);
	outportb(Qbase[qpnt] + QAINPAR, Qpars[qpddw]);
    }
}



void 
QADisss(int qdevice, float qisss)
{
    unsigned qpnt;

    qpnt = Qpointer(qdevice, "QADisss");
    if (qisss < 0.05 || qisss > 12.50)
	Qerror("Illegal ISSS rate specified", "QADisss");
    outport(Qbase[qpnt] + QSEPERATE, (int) (257.0 - qisss * 10.0));
}


void 
QADlast(int qdevice, unsigned qpddw)
{
    unsigned qpnt, qi;


    qpnt = Qpointer(qdevice, "QADgain");
    if (qpddw < 1 || qpddw > 16)
	Qerror("Illegal PDDW specified", "QADlast");
    Qchans = qpddw;
    qlaster = qpddw;
    for (qi = 1; qi <= 16; qi++) {
	Qpars[qi] = Qpars[qi] & 0x0B;
	if (qi == qpddw)
	    Qpars[qi] = Qpars[qi] | QLAST;
	outportb(Qbase[qpnt] + QPDDWSEL, qi - 1);
	outportb(Qbase[qpnt] + QAINPAR, Qpars[qi]);
    }
    if (qpddw == 1)
	QADisss(qdevice, 0.4);
    else if (Qhard[qpnt] == 1)
	QADisss(qdevice, 5.0);
    else
	QADisss(qdevice, 2.0);
}



void 
QADlevel(int qdevice, int qvalue)
{
    unsigned qpnt;

    qpnt = Qpointer(qdevice, "QADgain");
    outportb(Qbase[qpnt] + QCOMVAL, ((qvalue >> 8) ^ 0x80) & 0xFF);
}



unsigned 
QADoverflow(int qdevice)
{
    unsigned qpnt;


    qpnt = Qpointer(qdevice, "QADlevel");
    return (inportb(Qbase[qpnt] + QOVERFLOW));
}



void 
QADsamhold(int qdevice, unsigned qshmode)
{
    unsigned qpnt, qi;


    qpnt = Qpointer(qdevice, "QADsam_hold");
    if (qshmode < 0 || qshmode > 2)
	Qerror("Illegal sample/hold mode flag specified", "QADsam_hold");
    Qcmd[qpnt] = (Qcmd[qpnt] & 0xBF) | (qshmode << 6);
}




/*******    QDA1 SPECIFIC PROCEDURES     *********/





void 
QDA1reps(int qdevice, unsigned qrepa, unsigned qrepb)
{
    unsigned qpnter;

    if (qdevice != QDA1 && qdevice != QDA1x)
	Qerror("Illegal device specified", "QDA1reps");
    qpnter = Qpointer(qdevice, "QDA1reps");
    if (qrepa > 255 || qrepb > 255)
	Qerror("Rep-count specified is out of range", "QDA1reps");
    if (qrepa == 0) {
	QDA1cmd[qpnter] = QDA1cmd[qpnter] | QDA1_MANA;
    } else {
	QDA1cmd[qpnter] = QDA1cmd[qpnter] & (QDA1_MANA ^ 0xFF);
	QDA1repa[qpnter] = (255 - qrepa) & 0x0FF;
    }
    if (qrepb == 0) {
	QDA1cmd[qpnter] = QDA1cmd[qpnter] | QDA1_MANB;
    } else {
	QDA1cmd[qpnter] = QDA1cmd[qpnter] & (QDA1_MANB ^ 0xFF);
	QDA1repb[qpnter] = (255 - qrepb) & 0x0FF;
    }
}


void 
QDA1setsubs(int qdevice, unsigned qns)
{
    int     qpnter;
    int     qb, qi, qj;


    if (qdevice != QDA1 && qdevice != QDA1x)
	Qerror("Illegal device specified", "QDA1setsubs");
    if (qns != 1 && qns != 2 && qns != 4 && qns != 8 && qns != 16 && qns != 32)
	Qerror("Illegal number of subsections specified", "QDA1setsubs");
    qpnter = Qpointer(qdevice, "QDA1setsubs");
    qb = Qbase[qpnter];
    outport(qb + QDA1_CLK, 0xFF00);
    outportb(qb + QDA1_CMD, QDA1_STPA + QDA1_STPB);
    outportb(qb + QDA1_SINK, 0);
    outportb(qb + QDA1_CMD, QDA1_LOAD + QDA1_MANA + QDA1_MANB);
    for (qi = 1; qi <= 32; qi++) {
	QDA1adda[qpnter][qi] = 0;
	QDA1addb[qpnter][qi] = 0;
	Qrammap[qpnter][qi] = 0;
    }
    QDA1subs[qpnter][1] = 0x0000;
    QDA1subs[qpnter][2] = 0x0000;

    if (qns == 2 || qns == 4 || qns == 8 || qns == 32)
	QDA1subs[qpnter][1] = 0x8000;

    if (qns == 16 || qns == 32)
	QDA1subs[qpnter][2] = 0x8000;

    outport(qb + QDA1_ADDA, QDA1subs[qpnter][1]);
    outport(qb + QDA1_ADDB, QDA1subs[qpnter][2]);
    outportb(qb + QDA1_CMD, QDA1_LOAD + QDA1_MANA + QDA1_MANB + QDA1_GOA + QDA1_GOB);
    outportb(qb + QDA1_SINK, 0);
    outport(qb + QDA1_DATA, 00 ^ 0x7FFF);
    outport(qb + QDA1_DATB, 00 ^ 0x7FFF);
    outportb(qb + QDA1_CMD, QDA1_STPA + QDA1_STPB);
    outportb(qb + QDA1_SINK, 0);
    QDA1cmd[qpnter] = 0;
    QDA1size[qpnter] = (int) (1.0 * 0x7FFE / qns);
    QDA1numsubs[qpnter] = qns;
}





/*******    QDA2 SPECIFIC PROCEDURES  *******/

long 
QDAmode(long qbuf, int qmde)
{
    qmde = (int) ((qmde / 100) + 1) << 12;
    return (qbuf + qmde);
}




long 
Q(int buf[])
{
    long    sgg;


    sgg = (long) (16.0 * FP_SEG(buf) + 1.0 * FP_OFF(buf));
    return (sgg);
}




void 
Qdeclare(int qid, long qpts, unsigned qtype, unsigned qdevice, char qname[])
{
    unsigned qpnt, qpg;
    unsigned qmax, qneed;
    unsigned qi;
    long    ql;


    if (qdevice != QDISK && (qdevice & 0xFFFE) != QDA1) {
	qpnt = Qpointer(qdevice, "Qdeclare");
	if (qid < 1 || qid > 250 || Qbufdev[qid] != 0)
	    Qerror("Illegal or duplicate QID specified", "Qdeclare");
	if (qtype < 1 || qtype > 20)
	    Qerror("Illegal buffer type specified", "Qdeclare");
	Qbufdev[qid] = qpnt;
	strcpy(Qbufnme[qid], qname);
	Qbuftyp[qid] = qtype;
	Qbufcnt[qid] = qpts;
	qpg = 0;
	qmax = Qsize[qpnt] << 3;
	if (qid > QCHAN_B) {
	    qpg = Qsize[qpnt] << 2;
	    qmax = qpg << 1;
	}
	qneed = (int) (1.0 * qpts * (qtype & 0x0F) / 131072.0 + 1.0);
	do {
	    do {
		qpg++;
	    }
	    while (Qrammap[qpnt][qpg] > 0 && qpg < qmax);
	    if (qpg == qmax)
		Qerror("Not enough RAM for new buffer", "Qdeclare");
	    if ((qpg + qneed) > qmax)
		Qerror("Not enough RAM for new buffer", "Qdeclare");
	    Qbufpag[qid] = qpg;
	    for (qi = 0; qi < qneed; qi++)
		if (Qrammap[qpnt][qpg + qi] != 0)
		    Qbufpag[qid] = 100;
	}
	while (Qbufpag[qid] >= 100);
	do {
	    Qrammap[qpnt][qpg] = 1;
	    qpg++;
	    qneed--;
	}
	while (qneed != 0);
    }
    if (qdevice == QDISK) {
	Qbufdev[qid] = QFLAG;
	strcpy(Qbufnme[qid], qname);
	Qbuftyp[qid] = qtype;
	Qbufcnt[qid] = qpts;
    }
    if ((qdevice & 0xFFFE) == QDA1) {
	qpnt = Qpointer(qdevice, "Qdeclare");
	if (qid < 1 || qid > 250 || Qbufdev[qid] != 0)
	    Qerror("Illegal or duplicate QID specified", "Qdeclare");
	if (qtype != QINT && qtype != QWORD)
	    Qerror("Illegal buffer type specified for a QDA1", "Qdeclare");
	ql = qpts;
	if ((ql + 1) > QDA1size[qpnt]) {
	    Qerror("Buffer too large:  ", "Qdeclare");
	}
	Qbufdev[qid] = qpnt;
	strcpy(Qbufnme[qid], qname);
	Qbuftyp[qid] = qtype;
	Qbufcnt[qid] = qpts;
	if (qid >= QCHAN_B) {
	    qi = 0;
	    do {
		qi++;
		if (qi > QDA1numsubs[qpnt])
		    Qerror("No more buffers available on QDA1", "Qdeclare");
	    }
	    while (QDA1addb[qpnt][qi] != 0);
	    QDA1addb[qpnt][qi] = (int) ((32768.0 * qi / QDA1numsubs[qpnt]) - ql);
	    Qbufpag[qid] = qi;
	}
	if (qid < QCHAN_B) {
	    qi = 0;
	    do {
		qi++;
		if (qi > QDA1numsubs[qpnt])
		    Qerror("No more buffers available on QDA1", "Qdeclare");
	    }
	    while (QDA1adda[qpnt][qi] != 0);
	    QDA1adda[qpnt][qi] = (int) (32768.0 * qi / QDA1numsubs[qpnt] - ql);
	    Qbufpag[qid] = qi;
	}
    }
}



void 
Qfind(int qbuf, int qpts)
{
    unsigned qi, qb;
    long    qj;
    unsigned qdevice;
    unsigned qnum;
    unsigned qx;
    long    qcnt;
    long    qsizer;
    int     qport1;
    int     qport2;


    if (qbuf >= 0x100) {
	Qerror("Inproper source buffer location specified", "Qfind");
    }
    if ((Qbuftyp[qbuf] & 0x0F) != QWORD) {
	Qerror("Inproper source buffer type (must be QINT or QWORD)", "Qfind");
    }
    qi = Qbufdev[qbuf];
    if (qi == 0)
	Qerror("Buffer not defined", "Qfind");
    qb = Qbase[qi];
    outportb(qb + QCOMMAND, QLOOP);
    qdevice = Qdev[qi];
    if (qpts == QALL)
	qsizer = Qbufcnt[qbuf];
    else
	qsizer = qpts;
    Qsetcnt(qdevice, qsizer, "Qfind");
    outportb(qb + QPAGE, Qbufpag[qbuf] | QRAGRUN);
    outportb(qb + QPAGE, Qbufpag[qbuf] | QRAGRUN);
    qcnt = 0;
    qport1 = qb + QRAM + 1;
    qport2 = qb + QINFO;
    qj = Qbufcnt[qbuf] + 1;
    do {
	qi = inportb(qport1);
	qx = inportb(qport2);
	qcnt++;
    }
    while ((qx & 0x01) == 0 || qcnt <= qj);
}


void 
Qforget(int qid)
{
    unsigned qpnt, qpg;
    unsigned qmax, qneed;
    unsigned qi, qj;

    if (qid < QALL || qid > 250)
	Qerror("Illegal buffer ID specified", "Qforget");
    if (qid == QALL) {
	for (qj = 1; qj <= 10; qj++)
	    for (qi = 0; qi <= 31; qj++)
		Qrammap[qj][qi] = 0;
	for (qi = 1; qj <= 250; qj++)
	    Qbufdev[qi] = 0;
	for (qi = 1; qi <= 32; qi++) {
	    QDA1adda[8][qi] = 0;
	    QDA1addb[8][qi] = 0;
	    QDA1adda[9][qi] = 0;
	    QDA1addb[9][qi] = 0;
	}
    } else {
	if (Qbufdev[qid] < 8) {
	    if (Qbufdev[qid] == 0)
		Qerror("Unknown buffer referenced", "Qforget");
	    qneed = (int) (1.0 * Qbufcnt[qid] * (Qbuftyp[qid] & 0x0F) / 131072.0 + 1.0);
	    for (qi = 0; qi < qneed; qi++)
		Qrammap[Qbufdev[qid]][Qbufpag[qid] + qi] = 0;
	}
	if (Qbufdev[qid] > 7)
	    if (qid >= QCHAN_B)
		QDA1addb[Qbufdev[qid]][Qbufpag[qid]] = 0;
	    else
		QDA1adda[Qbufdev[qid]][Qbufpag[qid]] = 0;
	Qbufdev[qid] = 0;
    }
}



/* sublemental MOVE procedures */


void 
QdevpcP(unsigned ioport, unsigned qseg, unsigned qofs, unsigned count)
{
    asm .8086;
    asm mov dx, ioport;
    asm mov es, qseg;
    asm mov di, qofs;
    asm mov bx, count;
  backup:
    asm in  ax, dx;
    asm mov es:[di], ax;
    asm inc di;
    asm inc di;
    asm dec bx;
    asm dec bx;
    asm jnz backup;
}


void 
QdevpcA(unsigned ioport, unsigned qseg, unsigned qofs, unsigned count)
{
    asm .286;
    asm mov dx, ioport;
    asm mov es, qseg;
    asm mov di, qofs;
    asm mov cx, count;
    asm push ds
            asm push es
            asm pop ds
            asm rep insw
            asm pop ds
}



void 
QpcdevP(unsigned qseg, unsigned qofs, unsigned ioport, unsigned count)
{
    asm .8086;
    asm mov dx, ioport;
    asm mov es, qseg;
    asm mov di, qofs;
    asm mov bx, count;
  backup:
  asm mov ax, es:[di];
    asm out dx, ax;
    asm inc di;
    asm inc di;
    asm dec bx;
    asm dec bx;
    asm jnz backup;
}


void 
QpcdevA(unsigned qseg, unsigned qofs, unsigned ioport, unsigned count)
{
    asm .286;
    asm mov dx, ioport;
    asm mov es, qseg;
    asm mov si, qofs;
    asm mov cx, count;
    asm rep outsw
}



void 
Qdevdisk(long qbuf1, long qbuf2, long qpnts)
{
    unsigned qi, qb;
    unsigned qs, qo, qp;
    unsigned ql, qh;
    int     qer;
    long    qn, qsizer;
    unsigned qdevice;
    long    qnum;
    long    qseg, qofs;
    char   *qhold;
    char    qc;
    long    qx;
    FILE   *qfile;


    qhold = (char *) malloc(Qbufsize);
    if (qhold == NULL)
	Qerror("Not enough heap memory for transfer buffer requested", "Qmove");
    if (qbuf1 < 0x100) {
	qi = Qbufdev[qbuf1];
	qb = Qbase[qi];
	if (qi == 0)
	    Qerror("Source buffer not defined", "Qmove");
	qdevice = Qdev[qi];
	qsizer = Qbufcnt[qbuf1] * (Qbuftyp[qbuf1] & 0x0F);
	Qsetcnt(qdevice, qsizer, "Qmove");
	outportb(qb + QPAGE, Qbufpag[qbuf1] | QRAGRUN);
	outportb(qb + QPAGE, Qbufpag[qbuf1] | QRAGRUN);
    } else {
	qi = Qpointer(qbuf1, "Qmove");
	qb = Qbase[qi];
	qdevice = Qdev[qi];
    }
    outportb(qb + QCOMMAND, QLOOP);
    ql = qb + QRAM;
    qsizer = qelem * qpnts;

    qfile = fopen(Qbufnme[qbuf2], "wb");
    if (qfile == NULL)
	Qerror("Trouble opening file", "Qmove");

    do {
	qx = qsizer;
	if (qx > Qbufsize)
	    qx = Qbufsize;
	qsizer = qsizer - qx;
	if (Qcomptype == Q86)
	    QdevpcP(ql, FP_SEG(qhold), FP_OFF(qhold), (int) qx);
	else
	    QdevpcA(ql, FP_SEG(qhold), FP_OFF(qhold), (int) qx);
	fwrite(qhold, 2, qx >> 1, qfile);
    }
    while (qsizer > 0);
    fclose(qfile);
    free(qhold);
}



void 
Qdiskdev(long qbuf1, long qbuf2, long qpnts)
{
    unsigned qi, qb;
    unsigned qs, qo, qp;
    unsigned ql, qh;
    int     qer;
    long    qn, qsizer;
    unsigned qdevice;
    long    qnum;
    long    qseg, qofs;
    char   *qhold;
    char    qc;
    long    qx;
    FILE   *qfile;


    qhold = (char *) malloc(Qbufsize);
    if (qhold == NULL)
	Qerror("Not enough heap memory for transfer buffer requested", "Qmove");
    if (qbuf2 < 0x100) {
	qi = Qbufdev[qbuf2];
	qb = Qbase[qi];
	if (qi == 0)
	    Qerror("Source buffer not defined", "Qmove");
	qdevice = Qdev[qi];
	qsizer = Qbufcnt[qbuf2] * (Qbuftyp[qbuf2] & 0x0F);
	Qsetcnt(qdevice, qsizer, "Qmove");
	outportb(qb + QPAGE, Qbufpag[qbuf2] | QRAGRUN);
	outportb(qb + QPAGE, Qbufpag[qbuf2] | QRAGRUN);
    } else {
	qi = Qpointer(qbuf2, "Qmove");
	qb = Qbase[qi];
	qdevice = Qdev[qi];
    }
    outportb(qb + QCOMMAND, QLOOP);
    ql = qb + QRAM;
    qsizer = qelem * qpnts;

    qfile = fopen(Qbufnme[qbuf1], "rb");
    if (qfile == NULL)
	Qerror("Trouble opening file", "Qmove");

    do {
	qx = qsizer;
	if (qx > Qbufsize)
	    qx = Qbufsize;
	qsizer = qsizer - qx;
	fread(qhold, 2, qx >> 1, qfile);

	if (Qcomptype == Q86)
	    QpcdevP(FP_SEG(qhold), FP_OFF(qhold), ql, (int) qx);
	else
	    QpcdevA(FP_SEG(qhold), FP_OFF(qhold), ql, (int) qx);
    }
    while (qsizer > 0);
    fclose(qfile);
    free(qhold);
}




void 
Qdskdsk(long qbuf1, long qbuf2, long qpnts)
{
    unsigned qi, qb;
    unsigned qs, qo, qp;
    unsigned ql, qh;
    int     qer;
    long    qn, qsizer;
    unsigned qdevice;
    long    qnum;
    long    qseg, qofs;
    char   *qhold;
    char    qc;
    long    qx;
    FILE   *qfile1;
    FILE   *qfile2;


    qhold = (char *) malloc(Qbufsize);
    if (qhold == NULL)
	Qerror("Not enough heap memory for transfer buffer requested", "Qmove");

    qfile1 = fopen(Qbufnme[qbuf1], "rb");
    if (qfile1 == NULL)
	Qerror("Trouble opening source file", "Qmove");

    qfile2 = fopen(Qbufnme[qbuf2], "wb");
    if (qfile2 == NULL)
	Qerror("Trouble opening destination file", "Qmove");

    qsizer = qelem * qpnts;

    do {
	qx = qsizer;
	if (qx > Qbufsize)
	    qx = Qbufsize;
	qsizer = qsizer - qx;

	fread(qhold, 2, qx >> 1, qfile1);
	fwrite(qhold, 2, qx >> 1, qfile2);

    }
    while (qsizer > 0);
    fclose(qfile1);
    fclose(qfile2);
    free(qhold);
}


void 
Qpcdisk(long qbuf1, long qbuf2, long qpnts)
{
    unsigned qi, qb;
    unsigned qs, qo, qp;
    unsigned ql, qh;
    int     qer;
    long    qn, qsizer;
    unsigned qdevice;
    long    qnum;
    long    qseg, qofs;
    char   *qhold;
    char    qc;
    long    qx;
    FILE   *qfile;


    qsizer = qelem * qpnts;

    qfile = fopen(Qbufnme[qbuf2], "wb");
    if (qfile == NULL)
	Qerror("Trouble opening file", "Qmove");

    do {
	qx = qsizer;
	if (qx > Qbufsize)
	    qx = Qbufsize;
	qsizer = qsizer - qx;
	qhold = MK_FP(qbuf1 >> 4, qbuf1 & 0x0F);
	fwrite(qhold, 2, qx >> 1, qfile);
	qbuf1 = qbuf1 + qx;
    }
    while (qsizer > 0);
    fclose(qfile);
}



void 
Qdiskpc(long qbuf1, long qbuf2, long qpnts)
{
    unsigned qi, qb;
    unsigned qs, qo, qp;
    unsigned ql, qh;
    int     qer;
    long    qn, qsizer;
    unsigned qdevice;
    long    qnum;
    long    qseg, qofs;
    char   *qhold;
    char    qc;
    long    qx;
    FILE   *qfile;


    qsizer = qelem * qpnts;

    qfile = fopen(Qbufnme[qbuf1], "rb");
    if (qfile == NULL)
	Qerror("Trouble opening file", "Qmove");

    do {
	qx = qsizer;
	if (qx > Qbufsize)
	    qx = Qbufsize;
	qsizer = qsizer - qx;
	qhold = MK_FP(qbuf2 >> 4, qbuf2 & 0x0F);
	fread(qhold, 2, qx >> 1, qfile);
	qbuf2 = qbuf2 + qx;
    }
    while (qsizer > 0);
    fclose(qfile);
}



void 
Qdevdev(long qbuf1, long qbuf2, long qpts)
{
    unsigned qb1, qb2;
    unsigned qi1, qi2;
    unsigned qdevice1;
    unsigned qdevice2;
    long    qsizer;
    long    qnum;
    char   *qhold;


    if (qbuf1 < 0x100) {
	qi1 = Qbufdev[qbuf1];
	qb1 = Qbase[qi1];
	if (qi1 == 0)
	    Qerror("Source buffer not defined", "Qmove");
	qdevice1 = Qdev[qi1];
	qsizer = Qbufcnt[qbuf1] * (Qbuftyp[qbuf1] & 0x0F);
	Qsetcnt(qdevice1, qsizer, "Qmove");
	outportb(qb1 + QPAGE, Qbufpag[qbuf1] | QRAGRUN);
	outportb(qb1 + QPAGE, Qbufpag[qbuf1] | QRAGRUN);
    } else {
	qi1 = Qpointer(qbuf1, "Qmove");
	qb1 = Qbase[qi1];
	qdevice1 = Qdev[qi1];
    }
    if (qbuf2 < 0x100) {
	qi2 = Qbufdev[qbuf2];
	qb2 = Qbase[qi2];
	if (qi2 == 0)
	    Qerror("Source buffer not defined", "Qmove");
	qdevice2 = Qdev[qi2];
	qsizer = Qbufcnt[qbuf2] * (Qbuftyp[qbuf2] & 0x0F);
	Qsetcnt(qdevice2, qsizer, "Qmove");
	outportb(qb2 + QPAGE, Qbufpag[qbuf2] | QRAGRUN);
	outportb(qb2 + QPAGE, Qbufpag[qbuf2] | QRAGRUN);
    } else {
	qi2 = Qpointer(qbuf2, "Qmove");
	qb2 = Qbase[qi2];
	qdevice2 = Qdev[qi2];
    }
    if (qdevice1 == qdevice2)
	Qerror("Inter-device transfers not supported", "Qmove");
    qsizer = qpts * qelem;
    outportb(qb1 + QCOMMAND, QLOOP);
    outportb(qb2 + QCOMMAND, QLOOP);
    qhold = (char *) malloc(Qbufsize);
    if (qhold == NULL)
	Qerror("Not enough heap memory for transfer buffer requested", "Qmove");
    do {
	qnum = qsizer;
	if (qnum > Qbufsize)
	    qnum = Qbufsize;
	if (Qcomptype == Q86)
	    QdevpcP(qb1 + QRAM, FP_SEG(qhold), FP_OFF(qhold), qnum);
	else
	    QdevpcA(qb1 + QRAM, FP_SEG(qhold), FP_OFF(qhold), qnum);
	if (Qcomptype == Q86)
	    QpcdevP(FP_SEG(qhold), FP_OFF(qhold), qb2 + QRAM, qnum);
	else
	    QpcdevA(FP_SEG(qhold), FP_OFF(qhold), qb2 + QRAM, qnum);
	qsizer = qsizer - qnum;
    }
    while (qsizer > 0);
    free(qhold);
}


void 
Qdevpc(long qbuf1, long qbuf2, long qpnts)
{
    unsigned qi, qb;
    unsigned qs, qo, qp;
    unsigned ql, qh;
    int     qer;
    long    qn, qsizer;
    unsigned qdevice;
    long    qnum;
    long    qseg, qofs;
    char    qc;
    long    qx;


    if (qbuf1 < 0x100) {
	qi = Qbufdev[qbuf1];
	qb = Qbase[qi];
	if (qi == 0)
	    Qerror("Source buffer not defined", "Qmove");
	qdevice = Qdev[qi];
	qsizer = Qbufcnt[qbuf1] * (Qbuftyp[qbuf1] & 0x0F);
	Qsetcnt(qdevice, qsizer, "Qmove");
	outportb(qb + QPAGE, Qbufpag[qbuf1] | QRAGRUN);
	outportb(qb + QPAGE, Qbufpag[qbuf1] | QRAGRUN);
    } else {
	qi = Qpointer(qbuf1, "Qmove");
	qb = Qbase[qi];
	qdevice = Qdev[qi];
    }
    outportb(qb + QCOMMAND, QLOOP);
    ql = qb + QRAM;
    qsizer = qelem * qpnts;

    do {
	qx = qsizer;
	if (qx > Qbufsize)
	    qx = Qbufsize;
	qsizer = qsizer - qx;
	if (Qcomptype == Q86)
	    QdevpcP(ql, qbuf2 >> 4, qbuf2 & 0xF, qx);
	else
	    QdevpcA(ql, qbuf2 >> 4, qbuf2 & 0xF, qx);
	qbuf2 = qbuf2 + qx;
    }
    while (qsizer > 0);
}


void 
Qpcdev(long qbuf1, long qbuf2, long qpnts)
{
    unsigned qi, qb;
    unsigned qs, qo, qp;
    unsigned ql, qh;
    int     qer;
    long    qn, qsizer;
    unsigned qdevice;
    long    qnum;
    long    qseg, qofs;
    char    qc;
    long    qx;



    if (qbuf2 < 0x100) {
	qi = Qbufdev[qbuf2];
	qb = Qbase[qi];
	if (qi == 0)
	    Qerror("Source buffer not defined", "Qmove");
	qdevice = Qdev[qi];
	qsizer = Qbufcnt[qbuf2] * (Qbuftyp[qbuf2] & 0x0F);
	Qsetcnt(qdevice, qsizer, "Qmove");
	outportb(qb + QPAGE, Qbufpag[qbuf2] | QRAGRUN);
	outportb(qb + QPAGE, Qbufpag[qbuf2] | QRAGRUN);
    } else {
	qi = Qpointer(qbuf2, "Qmove");
	qb = Qbase[qi];
	qdevice = Qdev[qi];
    }
    outportb(qb + QCOMMAND, QLOOP);
    ql = qb + QRAM;
    qsizer = qelem * qpnts;

    do {
	qx = qsizer;
	if (qx > Qbufsize)
	    qx = Qbufsize;
	qsizer = qsizer - qx;
	if (Qcomptype == Q86)
	    QpcdevP(qbuf1 >> 4, qbuf1 & 0xF, ql, qx);
	else
	    QpcdevA(qbuf1 >> 4, qbuf1 & 0xF, ql, qx);
	qbuf1 = qbuf1 + qx;
    }
    while (qsizer > 0);
}


void 
QpcQDA1(long qbuf1, long qbuf2, long qpnts)
{
    unsigned qi, qb;
    unsigned qs, qo, qp;
    unsigned ql, qh;
    long    qn;
    unsigned qdevice;
    long    qcnt;
    long    qx;
    unsigned qadd;
    unsigned qdat;
    unsigned qchan;
    int    *qhold;

    if (qpnts != QALL) {
	if ((qpnts + 1) > QDA1size[Qbufdev[qbuf2]])
	    Qerror("Too many points specified", "Qmove");
	qs = (int) (32768.0 * Qbufpag[qbuf2] / QDA1numsubs[Qbufdev[qbuf2]] - qpnts);
	Qbufcnt[qbuf2] = qpnts;
	if (qbuf2 >= QCHAN_B)
	    QDA1addb[Qbufdev[qbuf2]][Qbufpag[qbuf2]] = qs;
	else
	    QDA1adda[Qbufdev[qbuf2]][Qbufpag[qbuf2]] = qs;
    }
    qcnt = Qbufcnt[qbuf2];
    qb = Qbase[Qbufdev[qbuf2]];
    outportb(qb + QDA1_CMD, QDA1_STPA + QDA1_STPB);
    outportb(qb + QDA1_SINK, 00);
    if (qbuf2 >= QCHAN_B) {
	qadd = qb + QDA1_ADDB;
	qdat = qb + QDA1_DATB;
	qchan = 2;
	qs = QDA1addb[Qbufdev[qbuf2]][Qbufpag[qbuf2]];
    } else {
	qadd = qb + QDA1_ADDA;
	qdat = qb + QDA1_DATA;
	qchan = 1;
	qs = QDA1adda[Qbufdev[qbuf2]][Qbufpag[qbuf2]];
    }
    qs = qs | QDA1subs[Qbufdev[qbuf2]][qchan];
    outport(qadd, qs);
    outportb(qb + QDA1_CMD, QDA1_GOA + QDA1_GOB + QDA1_MANA + QDA1_MANB + QDA1_LOAD);
    outportb(qb + QDA1_SINK, 00);
    qhold = MK_FP(qbuf1 >> 4, qbuf1 & 0xF);
    for (qi = 0; qi < qcnt; qi++)
	outport(qdat, qhold[qi] ^ 0x7FFF);
    outportb(qb + QDA1_CMD, QDA1_STPA + QDA1_STPB);
    outportb(qb + QDA1_SINK, 00);
}



void 
QdskQDA1(long qbuf1, long qbuf2, long qpnts)
{
    unsigned qi, qb;
    unsigned qs, qo, qp;
    unsigned ql, qh;
    long    qn;
    unsigned qdevice;
    long    qcnt;
    long    qx;
    unsigned qadd;
    unsigned qdat;
    unsigned qchan;
    int    *qhold;
    FILE   *qfile;

    if (qpnts != QALL) {
	if ((qpnts + 1) > QDA1size[Qbufdev[qbuf2]])
	    Qerror("Too many points specified", "Qmove");
	qs = (int) (1.0 * 0x8000 * Qbufpag[qbuf2] / QDA1numsubs[Qbufdev[qbuf2]] - qpnts);
	Qbufcnt[qbuf2] = qpnts;
	if (qbuf2 >= QCHAN_B)
	    QDA1addb[Qbufdev[qbuf2]][Qbufpag[qbuf2]] = qs;
	else
	    QDA1adda[Qbufdev[qbuf2]][Qbufpag[qbuf2]] = qs;
    }
    qcnt = Qbufcnt[qbuf2];
    qb = Qbase[Qbufdev[qbuf2]];
    outportb(qb + QDA1_CMD, QDA1_STPA + QDA1_STPB);
    outportb(qb + QDA1_SINK, 00);
    if (qbuf2 >= QCHAN_B) {
	qadd = qb + QDA1_ADDB;
	qdat = qb + QDA1_DATB;
	qchan = 2;
	qs = QDA1addb[Qbufdev[qbuf2]][Qbufpag[qbuf2]];
    } else {
	qadd = qb + QDA1_ADDA;
	qdat = qb + QDA1_DATA;
	qchan = 1;
	qs = QDA1adda[Qbufdev[qbuf2]][Qbufpag[qbuf2]];
    }
    qs = qs | QDA1subs[Qbufdev[qbuf2]][qchan];
    outport(qadd, qs);
    outportb(qb + QDA1_CMD, QDA1_GOA + QDA1_GOB + QDA1_MANA + QDA1_MANB + QDA1_LOAD);
    outportb(qb + QDA1_SINK, 00);

    qhold = (int *) malloc(qcnt << 1);
    if (qhold == NULL)
	Qerror("Not enough heap memory for transfer buffer requested", "Qmove");


    qfile = fopen(Qbufnme[qbuf1], "rb");
    if (qfile == NULL)
	Qerror("Trouble opening file", "Qmove");
    fread(qhold, 2, qcnt, qfile);

    for (qi = 0; qi < qcnt; qi++)
	outport(qdat, qhold[qi] ^ 0x7FFF);
    outportb(qb + QDA1_CMD, QDA1_STPA + QDA1_STPB);
    outportb(qb + QDA1_SINK, 00);
    fclose(qfile);
    free(qhold);
}



void 
QdevQDA1(long qbuf1, long qbuf2, long qpnts)
{
    unsigned qi, qb, qb2;
    unsigned ql, qh, qs;
    long    qn, qsizer;
    unsigned qdevice;
    long    qcnt;
    long    qx;
    unsigned qadd;
    unsigned qdat;
    unsigned qchan;
    int    *qhold;
    FILE   *qfile;

    if (qbuf1 < 0x100) {
	qi = Qbufdev[qbuf1];
	qb = Qbase[qi];
	if (qi == 0)
	    Qerror("Source buffer not defined", "Qmove");
	qdevice = Qdev[qi];
	qsizer = Qbufcnt[qbuf1] * (Qbuftyp[qbuf1] & 0x0F);
	Qsetcnt(qdevice, qsizer, "Qmove");
	outportb(qb + QPAGE, Qbufpag[qbuf1] | QRAGRUN);
	outportb(qb + QPAGE, Qbufpag[qbuf1] | QRAGRUN);
    } else {
	qi = Qpointer(qbuf1, "Qmove");
	qb = Qbase[qi];
	qdevice = Qdev[qi];
    }
    outportb(qb + QCOMMAND, QLOOP);
    if (qpnts != QALL) {
	if ((qpnts + 1) > QDA1size[Qbufdev[qbuf2]])
	    Qerror("Too many points specified", "Qmove");
	qs = (int) (1.0 * 0x8000 * Qbufpag[qbuf2] / QDA1numsubs[Qbufdev[qbuf2]] - qpnts);
	Qbufcnt[qbuf2] = qpnts;
	if (qbuf2 >= QCHAN_B)
	    QDA1addb[Qbufdev[qbuf2]][Qbufpag[qbuf2]] = qs;
	else
	    QDA1adda[Qbufdev[qbuf2]][Qbufpag[qbuf2]] = qs;
    }
    qcnt = Qbufcnt[qbuf2];
    qb2 = Qbase[Qbufdev[qbuf2]];
    outportb(qb2 + QDA1_CMD, QDA1_STPA + QDA1_STPB);
    outportb(qb2 + QDA1_SINK, 00);
    if (qbuf2 >= QCHAN_B) {
	qadd = qb2 + QDA1_ADDB;
	qdat = qb2 + QDA1_DATB;
	qchan = 2;
	qs = QDA1addb[Qbufdev[qbuf2]][Qbufpag[qbuf2]];
    } else {
	qadd = qb2 + QDA1_ADDA;
	qdat = qb2 + QDA1_DATA;
	qchan = 1;
	qs = QDA1adda[Qbufdev[qbuf2]][Qbufpag[qbuf2]];
    }
    qs = qs | QDA1subs[Qbufdev[qbuf2]][qchan];
    outport(qadd, qs);
    outportb(qb2 + QDA1_CMD, QDA1_GOA + QDA1_GOB + QDA1_MANA + QDA1_MANB + QDA1_LOAD);
    outportb(qb2 + QDA1_SINK, 00);
    ql = qb + QRAM;
    for (qx = 1; qx <= qcnt; qx++)
	outport(qdat, inport(ql) ^ 0x7FFF);
    outportb(qb2 + QDA1_CMD, QDA1_STPA + QDA1_STPB);
    outportb(qb2 + QDA1_SINK, 00);
}



void 
Qmove(long qbuf1, long qbuf2, long qpnts)
{
    long    qsizer;
    char    qdflag;
    long    qdump;

    qelem = Qdeftype;
    qsizer = QALL;
    qdflag = 0;
    qdump = qdflag + qsizer;
    qdump++;


/*  decide on proper number of points  */
    if (qpnts == QALL) {
	if (qbuf1 < 0x100)
	    qpnts = Qbufcnt[qbuf1];
	else if (qbuf2 < 0x100)
	    qpnts = Qbufcnt[qbuf2];
	else
	    Qerror("Unspecified number of points in procedure call", "Qmove");
    }
/*  decide on an apropriate buffer TYPE  */
    if (qbuf1 < 0x100)
	qelem = Qbuftyp[qbuf1] & 0x0F;
    else if (qbuf2 < 0x100)
	qelem = Qbuftyp[qbuf2] & 0x0F;
    else if (qbuf1 < 0x1000)
	qelem = Qoldtype & 0x0F;
    else if (qbuf2 < 0x1000)
	qelem = Qoldtype & 0x0F;


/*  be sure QDA1 has not been specified as source  */
    if (qbuf1 < 0x100 && Qbufdev[qbuf1] != QFLAG && Qbufdev[qbuf1] > 7)
	Qerror("QDA1 buffer cannot be specified as a source buffer", "Qmove");


/*  check is QDA1 is destination  */
    if (qbuf2 < 0x100 && Qbufdev[qbuf2] != QFLAG && Qbufdev[qbuf2] > 7) {
	if (qbuf1 < 0x100 && Qbufdev[qbuf1] == QFLAG)
	    QdskQDA1(qbuf1, qbuf2, qpnts);
	else if (qbuf1 < 0x1000)
	    QdevQDA1(qbuf1, qbuf2, qpnts);
	if (qbuf1 > 0x1000)
	    QpcQDA1(qbuf1, qbuf2, qpnts);
	qbuf1 = 0;
	qbuf2 = 0;
    }
/*  check if disk to disk  */
    if (qbuf1 < 0x100 && Qbufdev[qbuf1] == QFLAG && qbuf2 < 0x100 && Qbufdev[qbuf2] == QFLAG) {
	Qdskdsk(qbuf1, qbuf2, qpnts);
    }
/*  disk to ??? transfer  */
    if (qbuf1 < 0x100 && qbuf1 != 0 && Qbufdev[qbuf1] == QFLAG) {
	if (qbuf2 > 0x1000) {	/* file to PC RAM   */
	    Qdiskpc(qbuf1, qbuf2, qpnts);
	} else {		/* disk to device */
	    Qdiskdev(qbuf1, qbuf2, qpnts);
	}
	qbuf1 = 0;
	qbuf2 = 0;
    }
/*   ??? to disk transfer  */
    if (qbuf2 < 0x100 && qbuf2 != 0 && Qbufdev[qbuf2] == QFLAG) {
	if (qbuf1 > 0x1000) {	/* PC RAM to DISK */
	    Qpcdisk(qbuf1, qbuf2, qpnts);
	} else {		/* device to disk */
	    Qdevdisk(qbuf1, qbuf2, qpnts);
	}
	qbuf1 = 0;
	qbuf2 = 0;
    }
/*   device to device buffer transfer  */
    if (qbuf1 < 0x1000 && qbuf2 < 0x1000 && (qbuf1 + qbuf2) != 0)
	Qdevdev(qbuf1, qbuf2, qpnts);


/*   device to PC buffer trasfer   */
    if (qbuf1 < 0x1000 && qbuf2 > 0x1000) {
	Qdevpc(qbuf1, qbuf2, qpnts);
    }
/*  PC buffer to device transfer  */
    if (qbuf1 > 0x1000 && qbuf2 < 0x1000) {
	Qpcdev(qbuf1, qbuf2, qpnts);
    }
}


void 
Qpurge(int qdevice)
{
    int     ql, qh;
    int     qi, qd;


    if (qdevice == QALL) {
	ql = 2;
	qh = 7;
    } else {
	ql = Qpointer(qdevice, "Qpurge");
	qh = ql;
    }
    for (qd = ql; qd <= qh; qd++) {
	for (qi = 1; qi <= 250; qi++)
	    if (Qbufdev[qi] == qd) {
		Qforget(qi);
		Qbufdev[qi] = qd;
	    }
	for (qi = 1; qi <= 250; qi++)
	    if (Qbufdev[qi] == qd) {
		Qbufdev[qi] = 0;
		Qdeclare(qi, Qbufcnt[qi], Qbuftyp[qi], Qdev[qd], Qbufnme[qi]);
	    }
    }
}



unsigned 
Qset(long qbuf)
{
    int     qi, qb, qdevice;

    if (qbuf > 0x100)
	Qerror("Buffer type not allowed for this procedure", "Qset");
    qi = Qbufdev[qbuf];
    qb = Qbase[qi];
    if (qi == 0)
	Qerror("Source buffer not defined", "Qmove");
    qdevice = Qdev[qi];
    if (qi < 8) {
	Qsetcnt(qdevice, 5000000, "Qset");
	outportb(qb + QPAGE, Qbufpag[qbuf] | QRAGRUN);
	outportb(qb + QPAGE, Qbufpag[qbuf] | QRAGRUN);
	return (qb + QRAM);
    } else
	Qerror("The QDA1 is not supported by this procedure", "Qset");
}


void 
Qsplit(long qbuffer, long qpts, unsigned qchans, qintptr qpoints[])
{
    unsigned qi, qj, qb;
    unsigned qk, ql;
    unsigned qdevice;
    unsigned qnum;
    unsigned qs, qo;
    int    *qsrc;
    int     qport;
    long    qsizer;
    long    qp;


    if (qbuffer > 0x1000) {
	qs = qbuffer >> 4;
	qo = qbuffer & 0x0F;
	qsrc = MK_FP(qs, qo);
	qp = 0;
	for (qi = 0; qi < qpts; qi++)
	    for (qj = 1; qj <= qchans; qj++) {
		qpoints[qj][qi] = qsrc[qp];
		qp = qp + 1;
		if (qp > 32000) {
		    qbuffer = qbuffer + qp;
		    qs = qbuffer >> 4;
		    qo = qbuffer & 0x0F;
		    qsrc = MK_FP(qs, qo);
		    qp = 0;
		}
	    }
    }
    if (qbuffer < 0x1000) {
	if (qbuffer < 0x100) {
	    qi = Qbufdev[qbuffer];
	    qb = Qbase[qi];
	    outportb(qb + QCOMMAND, QLOOP);
	    if ((Qbuftyp[qbuffer] & 0x0F) != QWORD)
		Qerror("Inproper source buffer type (must be QINT or QWORD)", "Qsplit");
	    if (qi == 0)
		Qerror("Source buffer not defined", "Qsplit");
	    outportb(qb + QPAGE, Qbufpag[qbuffer] | QRAGRUN);
	    outportb(qb + QPAGE, Qbufpag[qbuffer] | QRAGRUN);
	    qdevice = Qdev[qi];
	    qsizer = Qbufcnt[qbuffer] * (Qbuftyp[qbuffer] & 0x0F);
	    Qsetcnt(qdevice, qsizer, "Qsplit");
	} else {
	    qi = Qpointer(qbuffer, "Qsplit");
	    qb = Qbase[qi];
	    outportb(qb + QCOMMAND, QLOOP);
	    if (qchans > 1)
		for (qk = 1; qk < qchans; qk++)
		    ql = inportb(qb + QRAM + 1);
	}
	ql++;
	qport = qb + QRAM;
	for (qi = 0; qi < qpts; qi++)
	    for (qj = 1; qj <= qchans; qj++)
		qpoints[qj][qi] = inport(qport);
    }
}



/*******    INITIALIZATION     *********/

void 
Qinit(int qdevice)
{
    unsigned qi, qj, qk;
    unsigned qpnt;

    qpnt = Qpointer(qdevice, "Qinit");
    if (qdevice == QAD || qdevice == QADx) {
	Qcmd[qpnt] = 00;
	outportb(Qbase[qpnt] + QCOMMAND, 00);
	Qsrate(qdevice, 100.00, "Qinit");
	Qsetcnt(qdevice, 1000, "Qinit");
	Qtrig(qdevice, QEXT_LOW);
	QADsamhold(qdevice, QOFF);
	for (qi = 1; qi <= 16; qi++) {
	    QADchan(qdevice, qi, qi);
	    QADgain(qdevice, qi, Q00DB);
	    QADcompare(qdevice, qi, QOFF);
	    QADlevel(qdevice, 0);
	}
	QADlast(qdevice, 1);
	for (qi = 0; qi <= 31; qi++)
	    Qrammap[qpnt][qi] = 0;
	for (qi = 1; qi <= 250; qi++)
	    if (Qbufdev[qi] == qdevice)
		Qbufdev[qi] = 0;
    }
    if (qdevice == QDA || qdevice == QDAx) {
	Qcmd[qpnt] = 00;
	outportb(Qbase[qpnt] + QCOMMAND, 00);
	Qsrate(qdevice, 100.00, "Qinit");
	Qsetcnt(qdevice, 1000, "Qinit");
	Qtrig(qdevice, QEXT_LOW);
	for (qi = 0; qi <= 31; qi++)
	    Qrammap[qpnt][qi] = 0;
	for (qi = 1; qi <= 250; qi++)
	    if (Qbufdev[qi] == qdevice)
		Qbufdev[qi] = 0;
    }
    if (qdevice == QAP || qdevice == QAPx) {
	Qerror("Device not yet supported", "Qinit");
    }
}


void 
Qstart(void)
{
    int     qi, qj, qk;
    unsigned qb, qid;

    Qbufsize = QDEFBS;
    Qdeftype = QDEFAULTTYPE;
    Qcomptype = Q286;
    Qoldtype = Qdeftype;
    for (qi = 1; qi <= 10; qi++) {
	Qbase[qi] = 00;
	Qhard[qi] = 00;
	Qsize[qi] = 00;
	Qdev[qi] = 00;
    }
    for (qi = 1; qi <= 250; qi++)
	Qbufdev[qi] = 0;
    for (qi = 1; qi <= 16; qi++)
	Qpars[qi] = 0;
    for (qi = 2; qi <= 7; qi++) {
	qb = 0x300 + (qi << 4);
	if ((inportb(qb + QINFO) & 0x70) == (inportb(qb + QINFO) & 0x70) &&
	    (inportb(qb + QINFO) & 0x70) == (inportb(qb + QINFO) & 0x70) &&
	    (inportb(qb + QINFO) & 0x70) == (inportb(qb + QINFO) & 0x70))
	    qj = (inportb(qb + QINFO) >> 4) & 0x03;
	else
	    qj = 0;
	if (qj > 0) {
	    qj = qj << 1;
	    if (Qbase[qj] != 0)
		qj = qj + 1;
	    outportb(qb + QCOMMAND, 0);
	    qk = (inportb(qb + QINFO) >> 5) & 0x01;
	    Qhard[qj] = qk + 1;
	    qk = 0;
	    outportb(qb + QPAGE, 0x00);
	    outportb(qb + QPAGE, 0x00);
	    outport(qb + QRAM, 0x00);
	    if (inport(qb + QRAM) == 0x00)
		qk = 2;
	    outportb(qb + QPAGE, 0x10 + QRAGRUN);
	    outportb(qb + QPAGE, 0x10 + QRAGRUN);
	    outport(qb + QRAM, 0x00);
	    outportb(qb + QPAGE, 0x10 + QRAGRUN);
	    outportb(qb + QPAGE, 0x10 + QRAGRUN);
	    if (inport(qb + QRAM) == 0x00) {
		outportb(qb + QPAGE, 0x8 + QRAGRUN);
		outportb(qb + QPAGE, 0x8 + QRAGRUN);
		outport(qb + QRAM, 0x00);
		outportb(qb + QPAGE, 0x8 + QRAGRUN);
		outportb(qb + QPAGE, 0x8 + QRAGRUN);
		if (inport(qb + QRAM) == 0x00)
		    qk = 4;
	    }
	    if (qk > 0) {
		Qbase[qj] = qb;
		Qsize[qj] = qk;
		Qdev[qj] = ((qj >> 1) << 8) + (qj & 0x01);
	    } else
		Qdev[qj] = 0;
	}
    }
    for (qi = 2; qi <= 7; qi++) {
	if (Qbase[qi] > 0) {
	    qid = ((qi >> 1) << 8) + (qi & 0x01);
	    Qinit(qid);
	}
    }
    qi = (QDA1 >> 7) + (QDA1 & 0x01);
    Qbase[qi] = 0x300;
    Qbase[qi + 1] = 0x310;
    Qdev[qi] = QDA1;
    Qdev[qi + 1] = QDA1x;
}
