/* rap.c */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include "waved.h"
#include "video.h"
#include "iod.h"
#include "fk.h"

#define ESC     27
#define NLPC    12
#define NFFT    128
#define MAXNF	(NLPC/2)
#define TOP     0
#define BOT     (msgy - ln_sp)


char   *marktext();
int     getyes();
int     menu();
int     init_dac();
int     init_adc();
int     lastevent();
int     sigsize();
void    record();
void    levels();
void    areavt();
void    playseg();
void    rapmsg();
void    time_msg();
void    ordmrk();

extern int ln_sp;

static int xx1, xx2, yy1, yy2;
static int play_dev = -1;
static int play_mark_1[NWIND];
static int play_mark_2[NWIND];
static int firstplay = 1;

void
playquery(q)
{
    char    str[80], mt1[10], mt2[10];
    int     err, pm1, pm2, i;
    SIGNAL *s;

    s = &sig[aw];
    if (s->siglen == 0)
	return;

    if (firstplay) {
        for (i = 0; i < NWIND; i++)
            play_mark_1[i] = play_mark_2[i] = NOMARK;
        firstplay = 0;
    }
    pm1 = play_mark_1[aw];
    pm2 = play_mark_2[aw];
    if (q) {
	if (getwinmrk(" PLAY:", 0) == 0) {
	    clrmsg();
	    return;
	}
	play_mark_1[aw] = pm1 = seg_mark1;
	play_mark_2[aw] = pm2 = seg_mark2;
    } else if (pm1 == NOMARK || pm2 == NOMARK
        || s->sigmrk[pm1] == NOMARK || s->sigmrk[pm2] == NOMARK) {
	play_mark_1[aw] = pm1 = HOME_MARK;
	play_mark_2[aw] = pm2 = END_MARK;
    }
    pm1 = play_mark_1[aw];
    pm2 = play_mark_2[aw];
    strcpy(mt1, marktext(pm1));
    strcpy(mt2, marktext(pm2));

    if (play_dev != iod_num || iod_num < 0) {
	if ((err = init_iod()) != 0) {
	    rapmsg("init_iod", err);
	    return;
	}
	play_dev = iod_num;
    }

    if (s->sgbits != adbits || s->sigzer != adzero)
        if (sigcon(aw, adbits, adzero, 1) && scrollmode == 0)
            showdata();

    clrmsg();
    sprintf(str, " Playing from %s to %s ", mt1, mt2);
    putmsg(str);
    playseg(s->sigmrk[pm1], s->sigmrk[pm2]);
    clrmsg();
}

void
replay()
{
    ordmrk();
    playquery(0);
}

void
playseg(aaa, bbb)
long    aaa, bbb;
{
    int     err;
    long    nplay;
    SIGPTR  samppnt;
    SIGNAL *s;

    s = &sig[aw];
    samppnt = s->sigbuf + aaa;
    nplay = bbb - aaa;
#ifdef WIN32
    err = output_dac(samppnt, nplay, s->sgnchn, s->darate);
#else /* WIN32 */
    err = output_dac(samppnt, nplay, s->darate);
    if (err != 0) {
	rapmsg("output_dac", err);
    }
#endif /* WIN32 */
}

void
input()
{
    int     c;
    static char *mis[] = {
    	"Area", "Level", "Record", ""
    };
    
    c = menu("INPUT", mis);

    switch (c) {
    case 'A':
	areavt();
	break;
    case 'L':
	levels();
	break;
    case 'R':
	record();
	break;
    }
}

 /* record - from ADC */

void
record()
{
    int     err, c, nchn;
    SIGPTR  samppnt;
    SIGNAL *s;

    s = &sig[aw];
    if (s->siglen > 0) {
        putmsg(" DELETE current contents of window? (y/N) ");
        if (!getyes())
            return;
        sigdel(aw);
        showwindow(aw);
    }
    nchn = limit(1, s->sgnchn, MAX_CHAN);
    if (sigsize(aw, topsamp, nchn) == 0) {
	rapmsg("sigsize", 0);
	return;
    }
    if ((err = init_iod()) != 0) {
	rapmsg("init_iod", err);
	return;
    }
    putmsg(" Press any key to begin recording ... ");
    beep();
    c = getevent();
    clrmsg();
    if (c == ESC || c == MDL)
	return;
    putmsg(" Recording from ADC ... ");
    samppnt = s->sigbuf;
    err = input_adc(samppnt, topsamp, s->adrate);
    if (err != 0) {
	rapmsg("input_adc", err);
	s->siglen = 0;
	return;
    }
    s->siglen = adnpts;
    s->sgbits = adbits;
    s->sigzer = adzero;
    s->sgtime = time(NULL);
    s->darate = s->adrate;
    s->sigbuf[s->siglen] = s->sigzer;
    s->sigfmt = 0;

    curspnt = (s->siglen < scrszx) ? s->siglen / 2 : scrszx / 2;
    win[aw].winmrk1 = 0;
    win[aw].winmrk2 = sig[aw].siglen;
    blankmarks(aw);
    showdata();
    time_msg(adnpts, s->adrate);
}

/*****************************************************************************/

/* LEVELS to set the input levels for recording speech */

void
levels()
{
    int     err, i;
    int     val, vmax, v, yyy, lst;
    long    l0, l1, l2, ns;
    SIGNAL *s;

    s = &sig[aw];
    g_init();
    putmsg(" ADC Input Level    [ESC to exit]");
    lvly1 = TOP;
    lvly2 = BOT - 1;
    g_box(lvlx1, lvly1, lvlx2, lvly2, boxcol);
    g_bar(lvlx1 + 1, lvly1 + 1, lvlx2 - 1, lvly2 - 1, wnbgc);
    xx1 = lvlx1 + 8;
    yy1 = lvly1 + 1;
    xx2 = lvlx2 - 8;
    yy2 = lvly2 - 1;
    if ((err = init_iod()) != 0) {
	rapmsg("init_iod", err);
	makescrn();
	return;
    }
    ns = BUFSIZE;
    if (adbits > 12)
        vmax = (1 << (adbits - 5)) - 2;
    else
        vmax = (1 << (adbits - 1)) - 2;
    l0 = yy2;
    l1 = yy2 - yy1;
    l2 = vmax + 2;
    lst = yy2;
    err = input_adc(buffer, ns, s->adrate);	/* ignore the first set */
    for (;;) {
	err = input_adc(buffer, ns, s->adrate);
	if (adnpts != ns) 
	    break;
	val = 0;
	for (i = 0; i < ns; i++) {
	    v = buffer[i] - adzero;
	    if (v > val)
		val = v;
	    else if (-v > val)
		val = -v;
	}
       if (adbits > 12)
            val /= 16;
	yyy = l0 - (val * l1) / l2;
	if (lst < yyy)
	    g_bar(xx1, lst, xx2, yyy, wnbgc);
	else
	    g_bar(xx1, yyy, xx2, lst, wavcol);
	if (val >= vmax)
	    beep();
	lst = yyy;
        g_box(lvlx1, lvly1, lvlx2, lvly2, boxcol);
    }
    while (chkevent())
        (void) getevent();
    makescrn();
}

void
rapmsg(fun, err)
char   *fun;
int     err;
{
    char    msg[80];

    clrmsg();
    sprintf(msg, "Error in '%s', error code = %d ", fun, err);
    putmsg(msg);
    waitforkey();
    clrmsg();
}

void
time_msg(n, r)
long    n, r;
{
    char    msg[80];

    clrmsg();
    if (r <= 0)
	return;
    sprintf(msg, " Record time = %ld msec", (n * 1000) / r);
    putmsg(msg);
    waitforkey();
    clrmsg();
    return;
}

int
sigcon(int sn, int b, int z, int q)
{
    int     sb, sz, sf, sv, c, d;
    long    n;
    SIGPTR  pb;
    SIGNAL *s;

    s = &sig[sn];
    n = s->siglen;
    if (n <= 0 || b <= 0)
	return (0);
    sb = s->sgbits;
    sz = s->sigzer;
    if (z == sz && b == sb)
	return (0);
    if (q) {
	clrmsg();
	putmsg("Data format does not match.  Reformat data? (Y/N) ");
	c = getyes();
	if (!c) {
	    clrmsg();
	    return (0);
	}
    }
    clrmsg();
    putmsg("Rescaling the signal ...");

    pb = s->sigbuf;
    if (b < sb) {
	sf = sb - b;
	d = 1 << sf;
	while (n-- > 0) {
	    sv = (((*pb) - sz) / d) + z;
	    *pb++ = sv;
	}
    } else if (b > sb) {
	sf = b - sb;
	d = 1 << sf;
	while (n-- > 0) {
	    sv = (((*pb) - sz) * d) + z;
	    *pb++ = sv;
	}
    } else {
	d = z - sz;
	while (n-- > 0)
	    *pb++ += d;
    }

    s->sgbits = b;
    s->sigzer = z;
    s->sigbuf[s->siglen] = s->sigzer;

    clrmsg();
    return (1);
}

/*****************************************************************************/

#include "memory.h"

void
areavt()
{
    float  *x, *a, *rc, *r, alpha, sm, refl;
    float   ar[NLPC], sum, smx, scl;
    int     lx1, ly1, lx2, ly2;
    int     ax1, ay1, ax2, ay2;
    int     ax0, ay0, adx, ady;
    int     axx1[NLPC], axx2[NLPC], ayy[NLPC], aymn, aymx;
    int     c, err, i, nc = NLPC, nd = 10, ncp;
    int     val, vthr, vmax, v, yyy, lst, status;
    long    l0, l1, l2;
    long    ns = BUFSIZE;
    long    rate = 10000;

    g_init();
    putmsg(" ADC Input Level    &    Area of the 'vocal tract'   [ESC to exit]");
    ax1 = scrszx / 4;
    ay1 = TOP;
    ax2 = scrszx - 1;
    ay2 = scrszy / 2 + scrszy / 4;
    adx = (ax2 - ax1) / nc;
    ady = (ay2 - ay1) / nd;
    ax0 = ax1 + 1 + ((ax2 - ax1) - adx * nc) / 2;
    ay0 = ay1 + 1 + ((ay2 - ay1) - ady * nd) / 2;
    g_box(ax1, ay1, ax2, ay2, boxcol);
    g_bar(ax1 + 1, ay1 + 1, ax2 - 1, ay2 - 1, wnbgc);
    g_bar(ax0, ay0, ax0 + adx * nc - 1, ay0 + ady, wavcol);
    for (i = 0; i < nc; i++) {
	axx1[i] = ax0 + adx * i;
	axx2[i] = ax0 + adx * (i + 1) - 1;
	ayy[i] = ay0 + ady;
    }
    aymn = ay1 + 2;
    aymx = ay2 - 2;
    lx1 = 0;
    ly1 = TOP;
    lx2 = ax1 - 4;
    ly2 = BOT - 1;
    g_box(lx1, ly1, lx2, ly2, boxcol);
    g_bar(lx1 + 1, ly1 + 1, lx2 - 1, ly2 - 1, wnbgc);
    xx1 = lx1 + 8;
    yy1 = ly1 + 1;
    xx2 = lx2 - 8;
    yy2 = ly2 - 1;
    if ((err = init_iod()) != 0) {
	rapmsg("init_iod", err);
	makescrn();
	return;
    }
    ncp = nc + 1;
    x = (float *) MALLOC(ns, sizeof(float));
    rc = (float *) MALLOC(ncp * 3, sizeof(float));
    a = rc + ncp;
    r = a + ncp;
    if (rc == NULL || x == NULL) {
	putmsgwk("Problem allocating memory.");
	return;
    }
    if (adbits > 12)
        vmax = (1 << (adbits - 5)) - 2;
    else
        vmax = (1 << (adbits - 1)) - 2;
    vthr = vmax >> 4;
    l0 = yy2;
    l1 = yy2 - yy1;
    l2 = vmax + 2;
    lst = yy2;
    err = input_adc(buffer, ns, rate);
    for (;;) {
	err = input_adc(buffer, ns, rate);
	if (adnpts != ns) {
            c = lastevent();
	    if (c == ESC || c == MDL)
		break;
            c = getevent();
	    continue;
	}
	val = 0;
	for (i = 0; i < ns; i++) {
	    v = buffer[i] - adzero;
	    if (v > val)
		val = v;
	    else if (-v > val)
		val = -v;
	}
       if (adbits > 12)
            val >>= 4;
	yyy = l0 - (val * l1) / l2;
	if (lst < yyy)
	    g_bar(xx1, lst, xx2, yyy, wnbgc);
	else
	    g_bar(xx1, yyy, xx2, lst, wavcol);
	if (val >= vmax)
	    beep();
	lst = yyy;
	if (val >= vthr && val < vmax) {
	    x[0] = 0.0;
	    for (i = 1; i < ns; i++)
		x[i] = (float) buffer[i] - (float) buffer[i - 1];
	    status = autoc((int) ns, x, nc, a, &alpha, rc, r);
	    if (status == 0) {
		sm = 1.0;
                sum = 0;
                smx = 0;
		for (i = 0; i < nc; i++) {
		    refl = rc[nc - (i + 1)];
		    sm *= (float) ((1.0 + refl) / (1.0 - refl));
		    ar[i] = sm;
                    sum += sm;
                    if (smx < sm)
                        smx = sm;
		}
                scl = (ay2 - ay1 ) *
                    ((smx > 4 * sum / nc) ? (1 / smx) : (nc / sum / 4));
		for (i = 0; i < nc; i++) {
		    yyy = limit(aymn, (int) (ay0 + ar[i] * scl), aymx);
		    if (ayy[i] < yyy)
			g_bar(axx1[i], ayy[i], axx2[i], yyy, wavcol);
		    else
			g_bar(axx1[i], yyy + 1, axx2[i], ayy[i], wnbgc);
		    ayy[i] = yyy;
		}

	    }
	}
    }
    FREE(x);
    FREE(rc);
    while (chkevent())
        (void) getevent();
    makescrn();
}
