/* trans.c - transform */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "waved.h"
#include "memory.h"
#include "video.h"

#define NFFT	    512
#define FTSPC_FT    1
#define FTSPC_LP    2
#define MAX_LPC     99

#ifndef _MSC_VER
#define _hypot	hypot
#endif

int     autoc();
int     menu();
int     rfft();
int     sigsize();
void    calc_ft();
void    copy_rate();
void    fourier_transform(int, int, int);
void    lperr();
void    lpeseg();

extern int lpnco, lpwnt, lphfe;

static int ftspc = FTSPC_FT;
static int ftpad = 1;

void
transform()
{
    char    str[80];
    int     c;
    static char *mis[] = {"FFT", "LPE", "Set", ""};
    static char *mis1[] = {"FFT-only", "LPC-only", "Both", ""};
    
    c = menu("TRANSFORM", mis);

    switch (c) {
    case 'F':
	calc_ft();
	break;
    case 'L':
	lpeseg();
	break;
    case 'S':
        c = menu("FFT TRANSFORM", mis1);
        if (c == 'F')
            ftspc = FTSPC_FT;
        else if (c == 'L')
            ftspc = FTSPC_LP;
        else if (c == 'B')
            ftspc = FTSPC_FT | FTSPC_LP;
        sprintf(str, " Pad with NFFT zeros, power-of-2  = (%d) ", ftpad);
        putmsg(str);
        c = getevent();
        if (c >= '0' && c <= '9')
	    ftpad = c - '0';
        putmsg(" High-frequency emphasis? (y/N) ");
        lphfe = getyes();
	break;
    }
    clrmsg();
}

void
calc_ft()
{
    win_num = clpwin;
    if (getwinmrk(" FOURIER TRANSFORM:", 0)) {
	ordmrk();
        sigdel(win_num);
	fourier_transform(win_num, seg_mark1, seg_mark2);
        if (fsw == 0) {
            showwindow(clpwin);
        } else {
            fsw = 0;
            showall();
        }
    }
}

/* lpeseg - replace segment with linear prediction error */

void
lpeseg()
{
    win_num = clpwin;
    ordmrk();
    sigdel(win_num);
    copy_rate(win_num, aw);
    saveseg(win_num, HOME_MARK, END_MARK);
    lperr(win_num, HOME_MARK, END_MARK);
    showwindow(win_num);
}

static double
db(double x)
{
    x = (x > 0.0001) ? 20 * log10(x) : -80;
    return (x);
}

static double
dbmag(float *x, int nf)
{
    double sum, xrms;
    float  x_nf;
    int i;

    x[0] = (float) db(x[0]);
    x_nf = (float) db(x[1]);
    sum = x[0] + x_nf;
    for (i = 1; i < nf; i++) {
	xrms = _hypot(x[i * 2], x[i * 2 + 1]);
        x[i] = (float) db(xrms);
        sum += x[i];
    }
    x[nf] = x_nf;
  
    return (sum);
}

static void
window(wave, m, wt)
float   *wave;
short    m, wt;
{
    double    dp, ph, w, mc;
    short     i, k;
    static double tpi = 6.2831853;
    static double Rectangular[3] = {1.0};
    static double Hamming[3] = {0.53836, 0.46164};
    static double Blackman[3] = {0.42, 0.5, 0.08};
    static double *a[] = {Rectangular, Hamming, Blackman};
    static short ncoef[] = {1, 2, 3};

    if (wt < 0 || wt >= 5)
        return;
    dp = tpi / (m - 1);
    mc = (m - 1) / 2.0;
    for (i = 0; i < m; i++) {
	ph = (i - mc) * dp;
	w = a[wt][0];
	for (k = 1; k < ncoef[wt]; k++)
	    w = w + a[wt][k] * cos(ph * k);
	wave[i] = wave[i] * (float) w;
    }
}

void
get_data(float *x, SIGPTR s, int n, int z, int hfe, int wnt)
{
    int i;

    if (hfe) {
        x[0] = 0;
        for (i = 1; i < n; i++)
            x[i] = (float) s[i] - (float) s[i - 1];
    } else {
        for (i = 0; i < n; i++)
            x[i] = (float) (s[i] - z);
    }
    if (wnt > 0)
        window(x, n, wnt);
}

void
put_data(float *x, SIGPTR s, int n, int ic, int nc, double gn, double of)
{
    int i;
    double v;
 
    for (i = 0; i < n; i++) {
	v = (x[i] - of) * gn;
	v = floor(v + 0.5);     /* round to nearest integer */
	s[i * nc + ic] = (short) limit(-32768, v, 32767);
    }
}

void
fourier_transform(int num, int from, int to)
{
    char    str[80];
    double  sum_wav, sum_lpc, ftdbrf;
    int     nfft, i, zer, nchn, status, nco, ncp, nf, ichn = 0;
    float  *x;
    float  *a, *rc, *r, alpha;
    long    left, right, npts;
    SIGPTR  pbuf1, pbuf2;
    SIGNAL  *s1, *s2;
    static int nfft_max = 1 << (sizeof(int) * 8 - 2);

    if (num == aw)
	return;
    s1 = &sig[aw];
    s2 = &sig[num];
    if (s1 == NULL || s2 == NULL)
	return;
    left = s1->sigmrk[from];
    right = s1->sigmrk[to];
    npts = right - left;
    if (npts <= 0)
	return;
    for (nfft = 1; nfft < npts && nfft < nfft_max; nfft *= 2)
	continue;
    if (ftpad > 0 && ftpad < 16)
        for (i = 0; i < ftpad && nfft < nfft_max; i++)
            nfft *= 2;
    if (nfft < npts) {
	sprintf(str, " Too many points (%ld) for FFT", npts);
	putmsgwk(str);
	return;
    }
    nf = nfft / 2;
    nchn = ((ftspc & FTSPC_FT) ? 1 : 0) + ((ftspc & FTSPC_LP) ? 1 : 0);
    if (sigsize(num, (long) nf, nchn) == 0) {
	return;
    }
    x = (float *) MALLOC(nfft + 2, sizeof(float));
    if (x == NULL) {
	putmsgwk("Problem allocating memory for FFT.");
	return;
    }
    clrmsg();
    sprintf(str, " FT: npts=%ld, nfft=%d", npts, nfft);
    putmsg(str);

    zer = s1->sigzer;
    pbuf1 = s1->sigbuf + left;
    pbuf2 = s2->sigbuf;

    /* compute the waveform spectrum */
    get_data(x, pbuf1, (int) npts, zer, lphfe, lpwnt);
    for (i = npts; i < nfft; i++)
        x[i] = 0;
    (void) rfft(x, nfft);
    sum_wav = dbmag(x, nf);
    ftdbrf = db((double) npts) - 35.1;
    if (ftspc & FTSPC_FT) {        /* save the waveform spectrum */
        put_data(x, pbuf2, nf + 1, ichn++, nchn, ftdbsc, ftdbrf + ftdbof);
    }

    if (ftspc & FTSPC_LP) {        /* compute & save the LPC spectrum */
        nco = (lpnco > 0) ? lpnco : 2 + s1->adrate / 1000;
        ncp = nco + 1;
        rc = (float *) MALLOC(ncp * 3, sizeof(float));
        if (rc == NULL) {
            putmsgwk("Problem allocating memory for LPC.");
            FREE(x);
            return;
        }
        a = rc + ncp;
        r = a + ncp;
        get_data(x, pbuf1, (int) npts, zer, lphfe, lpwnt);
        status = autoc((int) npts, x, nco, a, &alpha, rc, r);
        if (status != 0) {
            putmsgwk("Singular autocorrelation matrix.");
            FREE(x);
            FREE(rc);
            return;
        }
        for (i = 0; i < nco; i++)
            x[i] = a[i];
        FREE(rc);
        while (i < nfft)
            x[i++] = 0;
        (void) rfft(x, nfft);
        sum_lpc = dbmag(x, nf);
        put_data(x, pbuf2, nf + 1, ichn++, nchn, -ftdbsc, 
            (sum_wav + sum_lpc) / nf - (ftdbrf + ftdbof));
    }
    FREE(x);

    s2->siglen = nf;
    s2->sgnchn = nchn;
    s2->sgtime = time(NULL);
    s2->sgbits = 16;
    s2->sigzer = 0;
    s2->adrate = s1->adrate;
    s2->darate = 1;
    sprintf(s2->sigfnm, "[FFT of w%d(%ld:%ld)]", aw, left, right);
    s2->sigfmt = 1;
    blankmarks(num);
    win[num].winmrk1 = 0;
    win[num].winmrk2 = nf;
}

void
lperr(int num, int from, int to)
{
    char    str[80];
    int     zer, nc;
    float   rc[MAX_LPC+1];
    long    aaa, bbb, i, npt;
    SIGPTR  buf;
    SIGNAL *s;

    clrmsg();
    s = &sig[num];
    if (s == NULL)
	return;
    aaa = s->sigmrk[from];
    bbb = s->sigmrk[to];
    if (aaa == NOMARK || bbb == NOMARK)
	return;

    get_response("LP ERROR: Number of coefficients? ", str, 80);

    nc = atoi(str);
    if (nc <= 0 || nc > MAX_LPC)
	return;
    for (i = 0; i <= nc; i++)
	rc[i] = 0.0;
    buf = s->sigbuf + aaa;
    zer = s->sigzer;
    npt = bbb - aaa;
    for (i = 0; i < npt; i++)
	buf[i] = streak(buf[i] - zer, nc, rc) + zer;
    sprintf(s->sigfnm, "[LPE of w%d(%ld:%ld)]", aw, aaa, bbb);
}
