/* waved.c */

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "waved.h"
#include "video.h"
#include "memory.h"
#include "iod.h"
#include "fk.h"
#include "version.h"

#ifdef WIN32 
#define _strlwr	    strlwr
#endif
#ifdef DJC 
#define _strlwr	    strlwr
#define _strnicmp   strncasecmp
#endif

#define Ctrl_C      3
#define ESC         27
#define MAXNFN      3

char   *getenv();
int     adc_init();
int     adc_val();
int     getevent();
int     guess_file_type();
int     menu();
long    atol();
void    leave();
void    makescrn();
void    ins_mark();
void    ins_new_mark();
void    del_mark();
void    clrmsg();
void    putmsg();
void    clrsts();
void    putsts();
void    putmsgwk();
void    waitforkey();
void    read_args();
void    chkenv();
void    setup();
void    mouse_init();
void    showall();
void    goto_mark();
void    goto_samp();
void    mouse_left();
void    mouse_right();
void    calcquery();
void    editquery();
void    filequery();
void    modquery();
void    playquery();
void    replay();
void    viewquery();
void    windquery();
void    transform();
void    helptext();
void    showhelp();
void    usage();
void    loadfile();
void    sigmsg();
void    input();
void    init_wind();
void    showwindow();
void    showwindow_sm();
void    showdata();
void    showscrollbox();
void    showscroll_sm();
void    update_cursor();
void    update_mark();
void    chaw();
void    g_init();
void    g_close();
void    g_settextclr();
void    g_text();
void    g_text_dif();
void    g_bar();
void    g_gettxt();
void    zoom_in();
void    zoom_out();
void    verify_cl_files();
void    load_cl_files();
void    pan_left();
void    pan_right();
void    window_scroll();
void    window_max();

double  ftdbof = 60, ftdbsc = 2 * 32767.0 / 120;
int     aw = 1;
int     thru = 0;
int     fsw = 1;
int     lsw = NWIND - 1;
int     nsw = NWIND - 1;
int     clpwin = 0;
int     monoflg = 0;
int     no_header = 0;
int     win_num, seg_mark1, seg_mark2;
int     scroff;
int     scrollbox = 0;
int     scrollmode = 0;
int     maximize = 0;
int     siginfmsg = 0;
int     fmwnt = 1;		/* frequency moment window type (0=R,1=H,2=B) */
int     fmhfe = 1;		/* frequency moment high-freq. emphasis */
int     lpnco = 0;		/* number of linear prediction coefficients */
int     lpwnt = 1;		/* linear prediction window type (0=R,1=H,2=B) */
int     lphfe = 1;		/* linear prediction high-freq. emphasis */
long    curspnt = 0;
long    cursmin = 0;
long    cursmax = 0;
long    topsamp = 10000;	/* default maximum sample */
long    def_rate = 10000;	/* default ADC sample rate */
long    head_size = -1;         /* specified header size */
short   cpu_lsb = 0;
short   buffer[BUFSIZE];

WINDOW  win[NWIND];
SIGNAL  sig[NWIND];

static int nclfn = 0;
static char clfn[NWIND][PATHSZ];
static char lstmsg[80] = {0};
static char lststs[80] = {0};

void
verify_cl_files()
{
    char    tfn[FNSZ];
    int     w, n;

    n = 0;
    for (w = 0; w < nclfn; w++) {
	cdcpfn(tfn, clfn[w]);
	if (guess_file_type(tfn) < 0) {
#ifndef WIN32
	    fprintf(stderr, "can't open %s\n", clfn[w]);
#endif /* WIN32 */
	    strcpy(clfn[w], clfn[w + 1]);
	    w--;
	    nclfn--;
	    n++;
	}
    }
#ifndef WIN32
    if (n > 0) {
	fprintf(stderr, "Press any key to continue\n");
	(void) getevent();
    }
#endif /* WIN32 */
}

void
load_cl_files()
{
    char    tfn[FNSZ];
    int     n, w;

    lsw = fsw;
    if (nclfn <= 0) {
	showdata();
    } else {
	n = limit(1, nclfn, MAXNSW);
	lsw = fsw + n - 1;
	siginfmsg = 0;
	for (w = 0; w < n; w++) {
	    chaw(fsw + w);
	    cdcpfn(tfn, clfn[w]);
	    loadfile(tfn, 0);
	}
	if (lsw > fsw)
	    chaw(fsw);
    }
    siginfmsg = 1;		/* allow signal info messages */
}

/* wave_edit - view and edit waveform data */

void
wave_edit()
{
    long    ci1, ci2, ci3, npt, curspos;
    int     key;
    static char *mis[] =
    {
	"File", "Edit", "Modify", "Trans", "Calc", "Input", "Play",
	"View", "Window", "Set", "Help", ""
    };

    curspos = curspnt;
    npt = cursmax - cursmin;
    ci1 = (npt < scrszx) ? 1 : npt / scrszx;
    ci2 = ci1 * 10;
    ci3 = ci1 * 50;

    key = menu("WavEd", mis);

    switch (key) {
    case Ctrl_C:                /* thru */
	thru = 1;
	break;
    case MDL:
	mouse_left();
	break;
    case MDR:
	mouse_right();
	break;
    case FK_Shift_Tab:
	chaw(fsw + ((aw - fsw) - 1 + nsw) % nsw);
	break;
    case FK_F1:
	showhelp();
	break;
    case FK_F2:
	sigmsg(aw);
	break;
    case FK_F3:
	scrollbox = !scrollbox;
	showscrollbox(aw);
	break;
    case FK_F4:
	window_max();
	break;
    case FK_F5:
    case '~':
	window_scroll();
	break;
    case FK_F6:
	win[aw].winmod = !win[aw].winmod;
	showwindow(aw);
	break;
    case FK_Home:
	curspnt = cursmin;
	break;
    case FK_Ctrl_Home:
	goto_samp(sig[aw].sigmrk[HOME_MARK]);
	break;
    case FK_Left_Arrow:
        curspnt -= ci1;
	break;
    case FK_Right_Arrow:
        curspnt += ci1;
	break;
    case FK_Up_Arrow:
	if (aw > fsw)
	    chaw(aw - 1);
	break;
    case FK_Down_Arrow:
	if ((aw - fsw) < nsw - 1)
	    chaw(aw + 1);
	break;
    case FK_End:
	curspnt = cursmax;
	break;
    case FK_Ctrl_End:
	goto_samp(sig[aw].sigmrk[END_MARK]);
	break;
    case FK_Ins:
	ins_mark();
	break;
    case '0':
	ins_new_mark();
	break;
    case FK_Del:
	del_mark();
	break;
    case FK_PgUp:
	pan_left();
	break;
    case FK_PgDn:
	pan_right();
	break;
    case FK_Ctrl_Left:
        curspnt -= ci3;
	break;
    case FK_Ctrl_Right:
        curspnt += ci3;
	break;
    case '6':
        curspnt += ci2;
	break;
    case '4':
        curspnt -= ci2;
	break;
    case 'C':
	calcquery();
	break;
    case 'E':
	editquery();
	break;
    case 'F':
	filequery();
	break;
    case 'G':
	goto_mark();
	break;
    case 'H':
	showhelp();
	break;
    case 'I':
	input();
	break;
    case 'M':
	modquery();
	break;
    case 'P':
	playquery(1);
	break;
    case 'S':
	setup();
	break;
    case 'T':
	transform();
	break;
    case 'V':
	viewquery();
	break;
    case 'W':
	windquery();
	break;
    case '+':
	zoom_in();
	break;
    case '-':
	zoom_out();
	break;
    case ' ':
    case '\n':
	clrmsg();
	break;
    case '\f':
	makescrn();
	break;
    case '\r':
    case '"':
	replay();
	break;
    case '\t':			/* Tab */
	chaw((fsw + ((aw - fsw) + 1) % nsw));
	break;
    case ESC:			/* Escape */
	break;
    default:
	return;
    }

    if (thru)
        return;

    if (curspnt != curspos) {
        if (curspnt <= cursmin && curspnt >= cursmax)
            update_cursor();
        else
            goto_samp(curspnt);
	showscrollbox(aw);
    }
}

void
killpgm()
{
    leave(9);
}

void
leave(s)
{
    int     i;

    g_close();
    stop_iod();
    for (i = 0; i < NWIND; i++)
	if (sig[i].sigbuf != NULL)
	    FREE(sig[i].sigbuf);
#ifndef WIN32	    
    fprintf(stderr, "\n%s\n%s\n%s\n", VER, CPR, DST);
    exit(s);
#endif /* WIN32 */
}

void
waitforkey()
{
    while (!chkevent())
	continue;
}

void
putmsgwk(char *msg)
{
    clrmsg();
    putmsg(msg);
    (void) getevent();
    clrmsg();
}

void
putmsg(char *msg)
{
    if (strcmp(lstmsg, msg)) {
	if (*lstmsg > 0)
	    clrmsg();
	g_settextclr(msfgc, msbgc);
	g_text(msgx, msgy, msg);
	strcpy(lstmsg, msg);
	(void) chkevent();          /* updates window */
    }
}

void
clrmsg()
{
    g_bar(0, msgy - ch_ht, scrszx, msgy - ch_ht + ln_sp, msbgc);
    *lstmsg = '\0';
}

void
putsts(char *sts)
{
    if (strcmp(lststs, sts)) {
	g_settextclr(stscol, msbgc);
	g_text_dif(stsx, stsy, sts, lststs);
	strcpy(lststs, sts);
    }
}

void
clrsts()
{
    g_bar(0, stsy - ch_ht, scrszx, stsy - ch_ht + ln_sp, msbgc);
    *lststs = '\0';
}

void
get_response(char *q, char *s, int n)
{
    putmsg(q);
    *s = '\0';
    if (g_getstr(s, n) == MDL)
	*s = '\0';
    clrmsg();
}

#ifndef WIN32
void
set_lsb()
{
    short   a = 1;
    char   *b = (char *) (&a);

    cpu_lsb = b[1] ? 1 : 0;
}

void
set_grf_def()
{
    switch (g_videotype()) {
	case PEGA:
	grfcrd = 4;
	break;
    case VGA_MONO:
    case VGA_COLOR:
	grfcrd = 3;
	break;
    case CGA:
	grfcrd = 2;
	break;
    case HERCULES:
	grfcrd = 1;
	break;
    case EGA_MONO:
    case EGA_COLOR:
    case EGA_HIGH:
	grfcrd = 0;
	break;
    default:
	grfcrd = NOGRAF;
    }
}
#endif /* WIN32 */

void
set_grf_dev()
{
#ifdef MSC
    if (grfcrd == NOGRAF) {
	printf("\n WavEd requires a graphics card (CGA, EGA, HGC, or VGA).\n");
	exit(1);
    }
#endif /* WIN32 */
    g_init();
}

void
read_args(ac, av)
int     ac;
char  **av;
{
    long n;
    
    while (ac > 1) {
	if (av[1][0] == '-') {
	    switch (av[1][1]) {
	    case 'g':
		(void) _strlwr(av[2]);
		if (ac > 2) {
		    if (strcmp(av[2], "ega") == 0)
		        grfcrd = 0;
		    else if (strcmp(av[2], "hgc") == 0)
		        grfcrd = 1;
		    else if (strcmp(av[2], "cga") == 0)
		        grfcrd = 2;
		    else if (strcmp(av[2], "vga") == 0)
		        grfcrd = 3;
		    else
		        grfcrd = NOGRAF;
		}
		ac--;
		av++;
		break;
	    case 'h':
		if (ac > 2) {
                    n = atol(av[2]);
                    if (n > 0) {
                        head_size = n;
                    }
		}
		av++;
		ac--;
		break;
	    case 'H':
		helptext();
		break;
	    case 'm':
		monoflg = 1;
		break;
	    case 'n':
                if (av[1][2] == 'h')
                    no_header = 1;
		break;
	    case 'p':
		if (ac > 2) {
#ifdef MSC
		    if (_strnicmp(av[2], "DT", 2) == 0) {
		        iod_num = 2;
		    } else if (_strnicmp(av[2], "TDT", 1) == 0) {
		        iod_num = 3;
		    } else if (_strnicmp(av[2], "MON", 1) == 0) {
		        iod_num = 4;
		    } else if (_strnicmp(av[2], "PIN", 1) == 0) {
		        iod_num = 5;
		    } else if (_strnicmp(av[2], "I", 1) == 0) {
		        iod_num = 0;
		    } else if (_strnicmp(av[2], "LPT", 3) == 0) {
		        set_dev_port(1, av[2][3]);
		    } else if (_strnicmp(av[2], "VMK", 3) == 0) {
		        set_dev_port(0, av[2][3]);
		    } else if (av[2][0] >= '0' && av[2][0] <= '9') {
		        set_hex_port(av[2]);
		    }
#endif /* MSC */
		    av++;
		    ac--;
		}
		break;
	    case 'r':
		if (ac > 2) {
                    n = atol(av[2]);
                    if (n > 0) {
                        def_rate = n;
                        topsamp = n;
                    }
		}
		av++;
		ac--;
		break;
	    default:
		usage();
	    }
	} else if (nclfn < MAXNSW) {
	    strcpy(clfn[nclfn++], av[1]);
	}
	ac--;
	av++;
    }
}

void
chkenv()
{
    char    eb[256];
    char   *eav[32], *evs;
    int     eac;
    static char *evn = "WAVED";

    if ((evs = getenv(evn)) == NULL)
	return;
    fprintf(stderr, "%s=%s\n", evn, evs);
    strcpy(eb, evs);
    eav[0] = evn;
    eac = 1;
    for (evs = eb; *evs != '\0'; evs++) {
	while (*evs == ' ' || *evs == '\t')	/* skip blanks */
	    evs++;
	if (*evs == '\0')	/* quit if EOS */
	    break;
	eav[eac++] = evs;	/* point to arg */
	while (*evs != ' ' && *evs != '\t' && *evs != '\0')	/* skip arg */
	    evs++;
	if (*evs == '\0')	/* quit if EOS */
	    break;
	*evs = '\0';		/* terminate arg */
    }
    if (eac > 1)
	read_args(eac, eav);
}

void
showdata()
{
    if (scrollmode == 0) {
	showwindow(aw);
    } else {
	showwindow_sm();
    }
}

void
showall()
{
    int     i;

    showdata();
    for (i = fsw; i <= lsw; i++)
	if (i != aw)
	    showwindow(i);
}

/* MAKESCRN draws the main screen of WAVED. */

void
makescrn()
{
    g_init();
    *lstmsg = '\0';
    *lststs = '\0';
    showall();
}

void
waved(int ac, char **av)
{
    read_args(ac, av);
    verify_cl_files();
    set_grf_dev();
    init_wind();
    mouse_init();
    load_cl_files();
    while (!thru)
	wave_edit();
    leave(0);
}

#ifndef WIN32
int
main(int ac, char **av)
{
#ifdef _MSC_VER
    (void) signal(SIGINT, (void (__cdecl *) (int)) killpgm);
#endif
#ifdef DJC
    (void) signal(SIGINT, killpgm);
#endif
    set_lsb();
    set_grf_def();
    chkenv();
    waved(ac, av);
    return (0);
}
#endif /* WIN32 */

